# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import numpy as np
import pytest
from astropop.math.physical import QFloat, UnitsError, units
from packaging import version

from astropop.testing import *

# Testing qfloat compatibility with Numpy ufuncs and array functions.


class TestQFloatNumpyArrayFuncs:
    """Test numpy array functions for numpy comatibility."""

    def test_error_not_handled(self):
        # handled QFloat must be ok
        qf = QFloat([1.0, 2.0, 3.0], [0.1, 0.2, 0.3], "m")
        res = np.sqrt(qf)

        # not handled QFloat must raise
        with pytest.raises(TypeError):
            np.frexp(qf)

    def test_error_only_call_method(self):
        qf = QFloat([1.0, 2.0, 3.0], [0.1, 0.2, 0.3], "m")
        with pytest.raises(TypeError):
            np.sin.at(qf, 0)

    def test_qfloat_np_append(self):
        qf1 = QFloat([1.0, 2.0, 3.0], [0.1, 0.2, 0.3], unit="m")
        qf2 = QFloat([1.0], [0.1], unit="km")
        qf3 = QFloat(1.0, 0.1, unit="km")
        qf4 = QFloat(0, 0)

        qf = np.append(qf1, qf1)
        assert_equal(qf.nominal, [1.0, 2.0, 3.0, 1.0, 2.0, 3.0])
        assert_equal(qf.std_dev, [0.1, 0.2, 0.3, 0.1, 0.2, 0.3])
        assert_equal(qf.unit, qf1.unit)

        # This should work and convert the unit.
        qf = np.append(qf1, qf2)
        assert_equal(qf.nominal, [1.0, 2.0, 3.0, 1000.0])
        assert_equal(qf.std_dev, [0.1, 0.2, 0.3, 100.0])
        assert_equal(qf.unit, qf1.unit)

        # Also this should work and convert the unit in the same way.
        qf = np.append(qf1, qf3)
        assert_equal(qf.nominal, [1.0, 2.0, 3.0, 1000.0])
        assert_equal(qf.std_dev, [0.1, 0.2, 0.3, 100.0])
        assert_equal(qf.unit, qf1.unit)

        # This should fail due to unit
        with pytest.raises(UnitsError):
            qf = np.append(qf1, qf4)

        # Testing with axis
        qf1 = QFloat([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
                     [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]], "m",)
        qf = np.append(qf1, QFloat([[8.0], [9.0]], [[0.8], [0.9]], "m"),
                       axis=1)
        assert_equal(qf.nominal, [[1.0, 2.0, 3.0, 8.0], [4.0, 5.0, 6.0, 9.0]])
        assert_equal(qf.std_dev, [[0.1, 0.2, 0.3, 0.8], [0.4, 0.5, 0.6, 0.9]])
        qf = np.append(qf1, QFloat([[7.0, 8.0, 9.0]], [[0.7, 0.8, 0.9]], "m"),
                       axis=0)
        assert_equal(qf.nominal, [[1.0, 2.0, 3.0],
                                  [4.0, 5.0, 6.0],
                                  [7.0, 8.0, 9.0]])
        assert_equal(qf.std_dev, [[0.1, 0.2, 0.3],
                                  [0.4, 0.5, 0.6],
                                  [0.7, 0.8, 0.9]])

    def test_qfloat_np_around(self):
        # single case
        qf = np.around(QFloat(1.02549, 0.135964))
        assert_equal(qf.nominal, 1)
        assert_equal(qf.std_dev, 0)

        qf = np.around(QFloat(1.02549, 0.135964), decimals=2)
        assert_equal(qf.nominal, 1.03)
        assert_equal(qf.std_dev, 0.14)

        # just check array too
        qf = np.around(QFloat([1.03256, 2.108645], [0.01456, 0.594324]),
                       decimals=2)
        assert_equal(qf.nominal, [1.03, 2.11])
        assert_equal(qf.std_dev, [0.01, 0.59])

    def test_qfloat_np_atleast_1d(self):
        # This function is not implemented, so should raise
        with pytest.raises(TypeError):
            np.atleast_1d(QFloat([1.0, 2.0], [0.1, 0.2], "m"))

    def test_qfloat_np_atleast_2d(self):
        # This function is not implemented, so should raise
        with pytest.raises(TypeError):
            np.atleast_2d(QFloat([1.0, 2.0], [0.1, 0.2], "m"))

    def test_qfloat_np_atleast_3d(self):
        # This function is not implemented, so should raise
        with pytest.raises(TypeError):
            np.atleast_3d(QFloat([1.0, 2.0], [0.1, 0.2], "m"))

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_broadcast(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_broadcast_to(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_ceil(self):
        raise NotImplementedError

    def test_qfloat_np_clip(self):
        arr = np.arange(10)
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.clip(qf, 2, 8)
        tgt = [2, 2, 2, 3, 4, 5, 6, 7, 8, 8]
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, arr * 0.1)
        assert_equal(qf.unit, res.unit)

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_columnstack(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_concatenate(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_copyto(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_cross(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_cumprod(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_cumsum(self):
        raise NotImplementedError

    def test_qfloat_np_delete(self):
        a = np.array([[1.0, 2.0, 3.0, 4.0],
                      [5.0, 6.0, 7.0, 8.0],
                      [9.0, 10.0, 11.0, 12.0]])
        qf = QFloat(a, a * 0.1, "m")
        res1 = np.delete(qf, 1, axis=0)
        assert_almost_equal(res1.nominal, [[1.0, 2.0, 3.0, 4.0],
                                           [9.0, 10.0, 11.0, 12.0]])
        assert_almost_equal(res1.std_dev, [[0.1, 0.2, 0.3, 0.4],
                                           [0.9, 1.0, 1.1, 1.2]])
        assert_equal(res1.unit, qf.unit)

        res2 = np.delete(qf, 1, axis=1)
        assert_almost_equal(res2.nominal, [[1.0, 3.0, 4.0],
                                           [5.0, 7.0, 8.0],
                                           [9.0, 11.0, 12.0]])
        assert_almost_equal(res2.std_dev, [[0.1, 0.3, 0.4],
                                           [0.5, 0.7, 0.8],
                                           [0.9, 1.1, 1.2]])
        assert_equal(res2.unit, qf.unit)

        res3 = np.delete(qf, np.s_[::2], 1)
        assert_almost_equal(res3.nominal,
                            [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]])
        assert_almost_equal(res3.std_dev,
                            [[0.2, 0.4], [0.6, 0.8], [1.0, 1.2]])
        assert_equal(res3.unit, qf.unit)

        res4 = np.delete(qf, [1, 3, 5])
        assert_almost_equal(res4.nominal,
                            [1.0, 3.0, 5.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0])
        assert_almost_equal(res4.std_dev,
                            [0.1, 0.3, 0.5, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2])
        assert_equal(res4.unit, qf.unit)

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_diff(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_dstack(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_ediff1d(self):
        raise NotImplementedError

    def test_qfloat_np_expand_dims(self):
        qf = QFloat(1.0, 0.1, "m")
        res1 = np.expand_dims(qf, axis=0)
        assert_almost_equal(res1.nominal, [1.0])
        assert_almost_equal(res1.std_dev, [0.1])
        assert_equal(res1.unit, qf.unit)
        assert_equal(res1.shape, (1,))

        qf = QFloat([1.0, 2.0], [0.1, 0.2], "m")
        res2 = np.expand_dims(qf, axis=0)
        assert_almost_equal(res2.nominal, [[1.0, 2.0]])
        assert_almost_equal(res2.std_dev, [[0.1, 0.2]])
        assert_equal(res2.unit, qf.unit)
        assert_equal(res2.shape, (1, 2))
        res3 = np.expand_dims(qf, axis=1)
        assert_almost_equal(res3.nominal, [[1.0], [2.0]])
        assert_almost_equal(res3.std_dev, [[0.1], [0.2]])
        assert_equal(res3.unit, qf.unit)
        assert_equal(res3.shape, (2, 1))

        if version.parse(np.version.full_version) >= version.parse('1.18.0'):
            res4 = np.expand_dims(qf, axis=(2, 0))
            assert_almost_equal(res4.nominal, [[[1.0], [2.0]]])
            assert_almost_equal(res4.std_dev, [[[0.1], [0.2]]])
            assert_equal(res4.unit, qf.unit)
            assert_equal(res4.shape, (1, 2, 1))

    def test_qfloat_np_flip(self):
        a = np.arange(8).reshape((2, 2, 2))
        qf = QFloat(a, a * 0.1, "m")

        res1 = np.flip(qf)
        assert_equal(res1.nominal, a[::-1, ::-1, ::-1])
        assert_equal(res1.std_dev, a[::-1, ::-1, ::-1] * 0.1)
        assert_equal(res1.unit, qf.unit)

        res2 = np.flip(qf, 0)
        assert_equal(res2.nominal, a[::-1, :, :])
        assert_equal(res2.std_dev, a[::-1, :, :] * 0.1)
        assert_equal(res2.unit, qf.unit)

        res3 = np.flip(qf, 1)
        assert_equal(res3.nominal, a[:, ::-1, :])
        assert_equal(res3.std_dev, a[:, ::-1, :] * 0.1)
        assert_equal(res3.unit, qf.unit)

        res4 = np.flip(qf, 2)
        assert_equal(res4.nominal, a[:, :, ::-1])
        assert_equal(res4.std_dev, a[:, :, ::-1] * 0.1)
        assert_equal(res4.unit, qf.unit)

        # just some static check
        qf = QFloat([[1, 2], [3, 4]], [[0.1, 0.2], [0.3, 0.4]], "m")

        res5 = np.flip(qf)
        assert_equal(res5.nominal, [[4, 3], [2, 1]])
        assert_equal(res5.std_dev, [[0.4, 0.3], [0.2, 0.1]])
        assert_equal(res5.unit, qf.unit)

        res6 = np.flip(qf, 0)
        assert_equal(res6.nominal, [[3, 4], [1, 2]])
        assert_equal(res6.std_dev, [[0.3, 0.4], [0.1, 0.2]])
        assert_equal(res6.unit, qf.unit)

        res7 = np.flip(qf, 1)
        assert_equal(res7.nominal, [[2, 1], [4, 3]])
        assert_equal(res7.std_dev, [[0.2, 0.1], [0.4, 0.3]])
        assert_equal(res7.unit, qf.unit)

    def test_qfloat_np_fliplr(self):
        a = np.arange(8).reshape((2, 2, 2))
        qf = QFloat(a, a * 0.1, "m")
        res = np.fliplr(qf)
        assert_equal(res.nominal, a[:, ::-1, :])
        assert_equal(res.std_dev, a[:, ::-1, :] * 0.1)
        assert_equal(res.unit, qf.unit)

        qf = QFloat([[1, 2], [3, 4]], [[0.1, 0.2], [0.3, 0.4]], "m")
        res = np.fliplr(qf)
        assert_equal(res.nominal, [[2, 1], [4, 3]])
        assert_equal(res.std_dev, [[0.2, 0.1], [0.4, 0.3]])
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_flipud(self):
        a = np.arange(8).reshape((2, 2, 2))
        qf = QFloat(a, a * 0.1, "m")
        res = np.flipud(qf)
        assert_equal(res.nominal, a[::-1, :, :])
        assert_equal(res.std_dev, a[::-1, :, :] * 0.1)
        assert_equal(res.unit, qf.unit)

        qf = QFloat([[1, 2], [3, 4]], [[0.1, 0.2], [0.3, 0.4]], "m")
        res = np.flipud(qf)
        assert_equal(res.nominal, [[3, 4], [1, 2]])
        assert_equal(res.std_dev, [[0.3, 0.4], [0.1, 0.2]])
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_insert(self):
        a = np.array([[1, 2], [3, 4], [5, 6]])
        qf = QFloat(a, a * 0.1, "m")

        res = np.insert(qf, 5, QFloat(999, 0.1, unit="m"))
        assert_almost_equal(res.nominal, [1, 2, 3, 4, 5, 999, 6])
        assert_almost_equal(res.std_dev, [0.1, 0.2, 0.3, 0.4, 0.5, 0.1, 0.6])
        assert_equal(res.unit, qf.unit)

        res = np.insert(qf, 1, QFloat(999, 0.1, unit="m"), axis=1)
        assert_almost_equal(res.nominal,
                            [[1, 999, 2], [3, 999, 4], [5, 999, 6]])
        assert_almost_equal(res.std_dev, [[0.1, 0.1, 0.2],
                                          [0.3, 0.1, 0.4],
                                          [0.5, 0.1, 0.6]])
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_mean(self):
        a = np.arange(8).reshape((2, 4))
        qf = QFloat(a, a * 0.1, "m")

        res = np.mean(qf)
        assert_almost_equal(res.nominal, np.mean(a))
        assert_almost_equal(res.std_dev, np.std(a)/np.sqrt(a.size))
        assert_equal(res.unit, qf.unit)

        res = np.mean(qf, axis=0)
        assert_almost_equal(res.nominal, np.mean(a, axis=0))
        assert_almost_equal(res.std_dev, np.std(a, axis=0)/np.sqrt(2))
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_moveaxis(self):
        arr = np.zeros((3, 4, 5))
        qf = QFloat(arr, unit='m')

        res = np.moveaxis(qf, 0, -1)
        assert_equal(res.shape, (4, 5, 3))
        assert_equal(res.unit, qf.unit)

        res = np.moveaxis(qf, -1, 0)
        assert_equal(res.shape, (5, 3, 4))
        assert_equal(res.unit, qf.unit)

        res = np.moveaxis(qf, (0, 1), (-1, -2))
        assert_equal(res.shape, (5, 4, 3))
        assert_equal(res.unit, qf.unit)

        res = np.moveaxis(qf, [0, 1, 2], [-1, -2, -3])
        assert_equal(res.shape, (5, 4, 3))
        assert_equal(res.unit, qf.unit)

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_nancumprod(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_nancumsum(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_nanprod(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_nansum(self):
        raise NotImplementedError

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_prod(self):
        raise NotImplementedError

    def test_qfloat_np_ravel(self):
        arr = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        tgt = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.ravel(qf)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_repeat(self):
        arr = np.array([1, 2, 3])
        tgt = np.array([1, 1, 2, 2, 3, 3])
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.repeat(qf, 2)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_reshape(self):
        arr = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        tgt = np.array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]])
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.reshape(qf, (2, 6))
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)
        assert_equal(res.shape, (2, 6))

    def test_qfloat_np_resize(self):
        arr = np.array([[1, 2], [3, 4]])
        qf = QFloat(arr, arr * 0.1, "m")

        shp = (2, 4)
        tgt = np.array([[1, 2, 3, 4], [1, 2, 3, 4]])
        res = np.resize(qf, shp)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)
        assert_equal(res.shape, shp)

        shp = (4, 2)
        tgt = np.array([[1, 2], [3, 4], [1, 2], [3, 4]])
        res = np.resize(qf, shp)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)
        assert_equal(res.shape, shp)

        shp = (4, 3)
        tgt = np.array([[1, 2, 3], [4, 1, 2], [3, 4, 1], [2, 3, 4]])
        res = np.resize(qf, shp)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)
        assert_equal(res.shape, shp)

        shp = (0,)
        # must fail due to empty array (not real)
        with pytest.raises(TypeError):
            res = np.resize(qf, shp)

    def test_qfloat_np_roll(self):
        arr = np.arange(10)
        qf = QFloat(arr, arr * 0.01, "m")

        off = 2
        tgt = np.array([8, 9, 0, 1, 2, 3, 4, 5, 6, 7])
        res = np.roll(qf, off)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.01)
        assert_equal(res.unit, qf.unit)

        off = -2
        tgt = np.array([2, 3, 4, 5, 6, 7, 8, 9, 0, 1])
        res = np.roll(qf, off)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.01)
        assert_equal(res.unit, qf.unit)

        arr = np.arange(12).reshape((4, 3))
        qf = QFloat(arr, arr * 0.01, "m")

        ax = 0
        off = 1
        tgt = np.array([[9, 10, 11], [0, 1, 2], [3, 4, 5], [6, 7, 8]])
        res = np.roll(qf, off, axis=ax)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.01)
        assert_equal(res.unit, qf.unit)

        ax = 1
        off = 1
        tgt = np.array([[2, 0, 1], [5, 3, 4], [8, 6, 7], [11, 9, 10]])
        res = np.roll(qf, off, axis=ax)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.01)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_rollaxis(self):
        arr = np.ones((3, 4, 5, 6))
        qf = QFloat(arr, arr * 0.01, "m")

        res = np.rollaxis(qf, 3, 1)
        assert_equal(res.shape, (3, 6, 4, 5))

        res = np.rollaxis(qf, 2)
        assert_equal(res.shape, (5, 3, 4, 6))

        res = np.rollaxis(qf, 1, 4)
        assert_equal(res.shape, (3, 5, 6, 4))

    def test_qfloat_np_round(self):
        # single case
        qf = np.round(QFloat(1.02549, 0.135964))
        assert_equal(qf.nominal, 1)
        assert_equal(qf.std_dev, 0)

        qf = np.round(QFloat(1.02549, 0.135964), decimals=2)
        assert_equal(qf.nominal, 1.03)
        assert_equal(qf.std_dev, 0.14)

        # just check array too
        qf = np.round(QFloat([1.03256, 2.108645], [0.01456, 0.594324]),
                      decimals=2)
        assert_equal(qf.nominal, [1.03, 2.11])
        assert_equal(qf.std_dev, [0.01, 0.59])

    def test_qfloat_np_rot90(self):
        arr = np.array([[0, 1, 2], [3, 4, 5]])
        b1 = np.array([[2, 5], [1, 4], [0, 3]])
        b2 = np.array([[5, 4, 3], [2, 1, 0]])
        b3 = np.array([[3, 0], [4, 1], [5, 2]])
        b4 = np.array([[0, 1, 2], [3, 4, 5]])
        qf = QFloat(arr, arr * 0.1, "m")

        for k in range(-3, 13, 4):
            res = np.rot90(qf, k=k)
            assert_equal(res.nominal, b1)
            assert_equal(res.std_dev, b1 * 0.1)
            assert_equal(res.unit, qf.unit)
        for k in range(-2, 13, 4):
            res = np.rot90(qf, k=k)
            assert_equal(res.nominal, b2)
            assert_equal(res.std_dev, b2 * 0.1)
            assert_equal(res.unit, qf.unit)
        for k in range(-1, 13, 4):
            res = np.rot90(qf, k=k)
            assert_equal(res.nominal, b3)
            assert_equal(res.std_dev, b3 * 0.1)
            assert_equal(res.unit, qf.unit)
        for k in range(0, 13, 4):
            res = np.rot90(qf, k=k)
            assert_equal(res.nominal, b4)
            assert_equal(res.std_dev, b4 * 0.1)
            assert_equal(res.unit, qf.unit)

        arr = np.arange(8).reshape((2, 2, 2))
        qf = QFloat(arr, arr * 0.1, "m")

        ax = (0, 1)
        tgt = np.array([[[2, 3], [6, 7]], [[0, 1], [4, 5]]])
        res = np.rot90(qf, axes=ax)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

        ax = (1, 2)
        tgt = np.array([[[1, 3], [0, 2]], [[5, 7], [4, 6]]])
        res = np.rot90(qf, axes=ax)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

        ax = (2, 0)
        tgt = np.array([[[4, 0], [6, 2]], [[5, 1], [7, 3]]])
        res = np.rot90(qf, axes=ax)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

        ax = (1, 0)
        tgt = np.array([[[4, 5], [0, 1]], [[6, 7], [2, 3]]])
        res = np.rot90(qf, axes=ax)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_shape(self):
        for shp in [(10,), (11, 12), (11, 12, 13)]:
            qf = QFloat(np.ones(shp), np.ones(shp), "m")
            assert_equal(np.shape(qf), shp)

    def test_qfloat_np_size(self):
        for shp in [(10,), (11, 12), (11, 12, 13)]:
            qf = QFloat(np.ones(shp), np.ones(shp), "m")
            assert_equal(np.size(qf), np.prod(shp))

    def test_qfloat_np_squeeze(self):
        arr = np.array([[[0], [1], [2]]])
        qf = QFloat(arr, arr * 0.01, "m")

        res = np.squeeze(qf)
        assert_equal(res.shape, (3,))
        assert_almost_equal(res.nominal, [0, 1, 2])
        assert_almost_equal(res.std_dev, [0, 0.01, 0.02])
        assert_equal(res.unit, qf.unit)

        res = np.squeeze(qf, axis=0)
        assert_equal(res.shape, (3, 1))
        assert_almost_equal(res.nominal, [[0], [1], [2]])
        assert_almost_equal(res.std_dev, [[0], [0.01], [0.02]])
        assert_equal(res.unit, qf.unit)

        with pytest.raises(ValueError):
            np.squeeze(qf, axis=1)

        res = np.squeeze(qf, axis=2)
        assert_equal(res.shape, (1, 3))
        assert_almost_equal(res.nominal, [[0, 1, 2]])
        assert_almost_equal(res.std_dev, [[0, 0.01, 0.02]])
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_sum(self):
        arr = np.ones(10).reshape((2, 5))
        qf = QFloat(arr, arr*0.1, "m")

        res = np.sum(qf)
        assert_equal(res.nominal, 10)
        assert_equal(res.std_dev, 0.1*np.sqrt(10))
        assert_equal(res.unit, qf.unit)

        res = np.sum(qf, axis=0)
        assert_equal(res.shape, [5])
        assert_almost_equal(res.nominal, np.ones(5)*2)
        assert_almost_equal(res.std_dev, np.ones(5)*np.sqrt(2)*0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_swapaxes(self):
        arr = np.array([[[0, 1], [2, 3]], [[4, 5], [6, 7]]])
        tgt = np.array([[[0, 4], [2, 6]], [[1, 5], [3, 7]]])
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.swapaxes(qf, 0, 2)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_take(self):
        arr = np.array([1, 2, 3, 4, 5])
        tgt = np.array([2, 3, 5])
        ind = [1, 2, 4]
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.take(qf, ind)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_tile(self):
        arr = np.array([0, 1, 2])
        qf = QFloat(arr, arr * 0.1)

        tile = 2
        tgt = np.array([0, 1, 2, 0, 1, 2])
        res = np.tile(qf, tile)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

        tile = (2, 2)
        tgt = np.array([[0, 1, 2, 0, 1, 2], [0, 1, 2, 0, 1, 2]])
        res = np.tile(qf, tile)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

        # More checking
        arr = np.array([[1, 2], [3, 4]])
        qf = QFloat(arr, arr * 0.1)

        tile = 2
        tgt = np.array([[1, 2, 1, 2], [3, 4, 3, 4]])
        res = np.tile(qf, tile)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

        tile = (2, 1)
        tgt = np.array([[1, 2], [3, 4], [1, 2], [3, 4]])
        res = np.tile(qf, tile)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    def test_qfloat_np_transpose(self):
        arr = np.array([[1, 2], [3, 4], [5, 6]])
        tgt = np.array([[1, 3, 5], [2, 4, 6]])
        qf = QFloat(arr, arr * 0.1, "m")

        res = np.transpose(qf)
        assert_almost_equal(res.nominal, tgt)
        assert_almost_equal(res.std_dev, tgt * 0.1)
        assert_equal(res.unit, qf.unit)

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_trunc(self):
        raise NotImplementedError


class TestQFloatNumpyUfuncs:
    """Test numpy array functions for numpy comatibility."""

    @pytest.mark.parametrize('func', [np.abs, np.absolute])
    def test_qfloat_np_absolute(self, func):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(-1.0, 0.1, 'm')
        qf3 = QFloat(-5.0, 0.1)
        qf4 = QFloat(-6)
        qf5 = QFloat([1, -1, 2, -2])

        assert_equal(func(qf1), QFloat(1.0, 0.1, 'm'))
        assert_equal(func(qf2), QFloat(1.0, 0.1, 'm'))
        assert_equal(func(qf3), QFloat(5.0, 0.1))
        assert_equal(func(qf4), QFloat(6))
        assert_equal(func(qf5), [1, 1, 2, 2])

        with pytest.raises(NotImplementedError):
            # out argument should fail
            func(qf1, out=[])

    def test_qfloat_np_add(self):
        qf1 = QFloat(2.0, 0.2, 'm')
        qf2 = QFloat(1.0, 0.1, 'm')
        qf3 = QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'm')
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.add(qf1, qf2)
        assert_equal(res.nominal, 3.0)
        assert_almost_equal(res.std_dev, 0.223606797749979)
        assert_equal(res.unit, units.Unit('m'))

        res = np.add(qf1, qf3)
        assert_equal(res.nominal, [3, 4, 5])
        assert_almost_equal(res.std_dev, [0.2236068, 0.28284271, 0.36055513])
        assert_equal(res.unit, units.Unit('m'))

        res = np.add(qf3, qf1)
        assert_equal(res.nominal, [3, 4, 5])
        assert_almost_equal(res.std_dev, [0.2236068, 0.28284271, 0.36055513])
        assert_equal(res.unit, units.Unit('m'))

        with pytest.raises(UnitsError):
            np.add(qf1, qf4)

        with pytest.raises(UnitsError):
            np.add(qf1, qf5)

        with pytest.raises(UnitsError):
            np.add(qf1, 1.0)

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.add(qf1, qf2, out=[])

    def test_qfloat_np_ceil(self):
        assert_equal(np.ceil(QFloat(1.5, 0.1, 'm')), QFloat(2.0, 0.0, 'm'))
        assert_equal(np.ceil(QFloat(-1.5, 0.1, 'm')), QFloat(-1.0, 0.0, 'm'))
        assert_equal(np.ceil(QFloat(0.2, 0.1, 'm')), QFloat(1.0, 0.0, 'm'))
        assert_equal(np.ceil(QFloat(-0.2, 0.1, 'm')), QFloat(0.0, 0.0, 'm'))

    @pytest.mark.parametrize('func', [np.divide, np.true_divide])
    def test_qfloat_np_divide(self, func):
        qf1 = QFloat(2.0, 0.2, 'm')
        qf2 = QFloat(1.0, 0.1, 'm')
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4], 'cm')
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = func(qf1, qf2)
        assert_equal(res.nominal, 2)
        assert_almost_equal(res.std_dev, 0.28284271)
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = func(qf1, qf3)
        assert_equal(res.nominal, [2, 1, 0.5])
        assert_almost_equal(res.std_dev, [0.28284271, 0.14142136, 0.07071068])
        assert_equal(res.unit, units.Unit('m/cm'))

        res = func(qf3, qf1)
        assert_equal(res.nominal, [0.5, 1, 2])
        assert_almost_equal(res.std_dev, [0.0707107, 0.1414214, 0.2828427])
        assert_equal(res.unit, units.Unit('cm/m'))

        res = func(qf1, qf4)
        assert_equal(res.nominal, 2.0)
        assert_almost_equal(res.std_dev, 0.28284271247461906)
        assert_equal(res.unit, units.Unit('m/s'))

        res = func(qf1, qf5)
        assert_equal(res.nominal, 2.0)
        assert_almost_equal(res.std_dev, 0.2)
        assert_equal(res.unit, units.Unit('m'))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            func(qf1, qf2, out=[])

    def test_qfloat_np_divmod(self):
        qf1 = QFloat(2.0, 0.2, 'm')
        qf2 = QFloat(1.0, 0.1, 'm')
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4], 'cm')
        qf4 = QFloat(1.0, 0.1, 's')

        res = np.divmod(qf1, qf2)
        assert_equal(res[0], np.floor_divide(qf1, qf2))
        assert_equal(res[1], np.remainder(qf1, qf2))

        res = np.divmod(qf1, qf3)
        assert_equal(res[0], np.floor_divide(qf1, qf3))
        assert_equal(res[1], np.remainder(qf1, qf3))

        res = np.divmod(qf3, qf1)
        assert_equal(res[0], np.floor_divide(qf3, qf1))
        assert_equal(res[1], np.remainder(qf3, qf1))

        res = np.divmod(qf1, qf4)
        assert_equal(res[0], np.floor_divide(qf1, qf4))
        assert_equal(res[1], np.remainder(qf1, qf4))

    def test_qfloat_np_exp(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.exp(qf1)
        assert_equal(res.nominal, np.exp(2.0))
        assert_almost_equal(res.std_dev, 0.2*np.exp(2.0))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.exp(qf2)
        assert_equal(res.nominal, np.exp(1.0))
        assert_almost_equal(res.std_dev, 0.1*np.exp(1.0))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.exp(qf3)
        assert_equal(res.nominal, np.exp(qf3.nominal))
        assert_almost_equal(res.std_dev, np.exp(qf3.nominal)*qf3.std_dev)
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='log is only defined for '
                           'dimensionless quantities.'):
            res = np.log(qf4)

        res = np.exp(qf5)
        assert_equal(res.nominal, np.exp(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_exp2(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.exp2(qf1)
        assert_equal(res.nominal, np.exp2(2.0))
        assert_almost_equal(res.std_dev, 0.2*np.exp2(2.0)*np.log(2))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.exp2(qf2)
        assert_equal(res.nominal, np.exp2(1.0))
        assert_almost_equal(res.std_dev, 0.1*np.exp2(1.0)*np.log(2))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.exp2(qf3)
        assert_equal(res.nominal, np.exp2(qf3.nominal))
        assert_almost_equal(res.std_dev,
                            np.exp2(qf3.nominal)*qf3.std_dev*np.log(2))
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='exp2 is only defined for '
                           'dimensionless quantities.'):
            res = np.exp2(qf4)

        res = np.exp2(qf5)
        assert_equal(res.nominal, np.exp2(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_expm1(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.expm1(qf1)
        assert_equal(res.nominal, np.expm1(2.0))
        assert_almost_equal(res.std_dev, 0.2*np.exp(2.0))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.expm1(qf2)
        assert_equal(res.nominal, np.expm1(1.0))
        assert_almost_equal(res.std_dev, 0.1*np.exp(1.0))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.expm1(qf3)
        assert_equal(res.nominal, np.expm1(qf3.nominal))
        assert_almost_equal(res.std_dev, np.exp(qf3.nominal)*qf3.std_dev)
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='log is only defined for '
                           'dimensionless quantities.'):
            res = np.log(qf4)

        res = np.expm1(qf5)
        assert_equal(res.nominal, np.expm1(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_floor(self):
        assert_equal(np.floor(QFloat(1.5, 0.1, 'm')), QFloat(1.0, 0.0, 'm'))
        assert_equal(np.floor(QFloat(-1.5, 0.1, 'm')), QFloat(-2.0, 0.0, 'm'))
        assert_equal(np.floor(QFloat(0.2, 0.1, 'm')), QFloat(0.0, 0.0, 'm'))
        assert_equal(np.floor(QFloat(-0.2, 0.1, 'm')), QFloat(-1.0, 0.0, 'm'))

    def test_qfloat_np_floor_divide(self):
        qf1 = QFloat(2.0, 0.2, 'm')
        qf2 = QFloat(1.0, 0.1, 'm')
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4], 'cm')
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.floor_divide(qf1, qf2)
        assert_equal(res.nominal, 2)
        assert_almost_equal(res.std_dev, 0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.floor_divide(qf1, qf3)
        assert_equal(res.nominal, [2, 1, 0])
        assert_almost_equal(res.std_dev, [0, 0, 0])
        assert_equal(res.unit, units.Unit('m/cm'))

        res = np.floor_divide(qf3, qf1)
        assert_equal(res.nominal, [0, 1, 2])
        assert_almost_equal(res.std_dev, [0, 0, 0])
        assert_equal(res.unit, units.Unit('cm/m'))

        res = np.floor_divide(qf1, qf4)
        assert_equal(res.nominal, 2)
        assert_almost_equal(res.std_dev, 0)
        assert_equal(res.unit, units.Unit('m/s'))

        res = np.floor_divide(qf1, qf5)
        assert_equal(res.nominal, 2)
        assert_almost_equal(res.std_dev, 0)
        assert_equal(res.unit, units.Unit('m'))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.floor_divide(qf1, qf2, out=[])

    def test_qfloat_np_hypot(self):
        qf1 = QFloat(3, 0.3, 'm')
        qf2 = QFloat(4, 0.4, 'm')
        qf3 = QFloat(3*np.ones((5, 5)), unit='m')
        qf4 = QFloat(4*np.ones((5, 5)), unit='m')

        res = np.hypot(qf1, qf2)
        assert_equal(res.nominal, 5)
        assert_almost_equal(res.std_dev, 0.36715119501371646)
        assert_equal(res.unit, units.Unit('m'))

        res = np.hypot(qf3, qf4)
        assert_equal(res.nominal, 5*np.ones((5, 5)))
        assert_almost_equal(res.std_dev, np.zeros((5, 5)))
        assert_equal(res.unit, units.Unit('m'))

        res = np.hypot(qf1, qf4)
        assert_equal(res.nominal, 5*np.ones((5, 5)))
        assert_almost_equal(res.std_dev, 0.18*np.ones((5, 5)))
        assert_equal(res.unit, units.Unit('m'))

        with pytest.raises(UnitsError):
            np.hypot(qf1, 1)

        with pytest.raises(UnitsError):
            np.hypot(qf1, QFloat(1, unit='s'))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.multiply(qf1, qf2, out=[])

    def test_qfloat_np_isfinit(self):
        assert_false(np.isfinite(QFloat(np.inf)))
        assert_false(np.isfinite(QFloat(np.nan)))
        assert_true(np.isfinite(QFloat(1)))
        assert_true(np.isfinite(QFloat(1, unit='m')))
        assert_true(np.isfinite(QFloat(1, 0.1, unit='m/s')))

        assert_equal(np.isfinite(QFloat([np.inf, np.nan, 1.0], unit='m')),
                     [False, False, True])

    def test_qfloat_np_isinf(self):
        assert_true(np.isinf(QFloat(np.inf)))
        assert_false(np.isinf(QFloat(np.nan)))
        assert_false(np.isinf(QFloat(1)))
        assert_false(np.isinf(QFloat(1, unit='m')))
        assert_false(np.isinf(QFloat(1, 0.1, unit='m/s')))

        assert_equal(np.isinf(QFloat([np.inf, np.nan, 1.0], unit='m')),
                     [True, False, False])

    def test_qfloat_np_isnan(self):
        assert_false(np.isnan(QFloat(np.inf)))
        assert_true(np.isnan(QFloat(np.nan)))
        assert_false(np.isnan(QFloat(1)))
        assert_false(np.isnan(QFloat(1, unit='m')))
        assert_false(np.isnan(QFloat(1, 0.1, unit='m/s')))

        assert_equal(np.isnan(QFloat([np.inf, np.nan, 1.0], unit='m')),
                     [False, True, False])

    def test_qfloat_np_log(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.log(qf1)
        assert_equal(res.nominal, np.log(2.0))
        assert_almost_equal(res.std_dev, 0.2/2.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log(qf2)
        assert_equal(res.nominal, np.log(1.0))
        assert_almost_equal(res.std_dev, 0.1/1.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log(qf3)
        assert_equal(res.nominal, np.log([1, 2, 4]))
        assert_almost_equal(res.std_dev,
                            qf3.std_dev/qf3.nominal)
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='log is only defined for '
                           'dimensionless quantities.'):
            res = np.log(qf4)

        res = np.log(qf5)
        assert_equal(res.nominal, np.log(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_log2(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.log2(qf1)
        assert_equal(res.nominal, np.log2(2.0))
        assert_almost_equal(res.std_dev, 0.2/(2.0*np.log(2)))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log2(qf2)
        assert_equal(res.nominal, np.log2(1.0))
        assert_almost_equal(res.std_dev, 0.1/np.log(2))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log2(qf3)
        assert_equal(res.nominal, np.log2([1, 2, 4]))
        assert_almost_equal(res.std_dev,
                            qf3.std_dev/(qf3.nominal*np.log(2)))
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='log2 is only defined for '
                           'dimensionless quantities.'):
            res = np.log2(qf4)

        res = np.log2(qf5)
        assert_equal(res.nominal, np.log2(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_log10(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.log10(qf1)
        assert_equal(res.nominal, np.log10(2.0))
        assert_almost_equal(res.std_dev, 0.2/(2.0*np.log(10)))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log10(qf2)
        assert_equal(res.nominal, np.log10(1.0))
        assert_almost_equal(res.std_dev, 0.1/np.log(10))
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log10(qf3)
        assert_equal(res.nominal, np.log10([1, 2, 4]))
        assert_almost_equal(res.std_dev,
                            qf3.std_dev/(qf3.nominal*np.log(10)))
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='log10 is only defined for '
                           'dimensionless quantities.'):
            res = np.log10(qf4)

        res = np.log10(qf5)
        assert_equal(res.nominal, np.log10(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_log1p(self):
        qf1 = QFloat(2.0, 0.2)
        qf2 = QFloat(1.0, 0.1)
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4])
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.log(qf1)
        assert_equal(res.nominal, np.log(2.0))
        assert_almost_equal(res.std_dev, 0.2/2.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log(qf2)
        assert_equal(res.nominal, np.log(1.0))
        assert_almost_equal(res.std_dev, 0.1/1.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        res = np.log(qf3)
        assert_equal(res.nominal, np.log([1, 2, 4]))
        assert_almost_equal(res.std_dev,
                            qf3.std_dev/qf3.nominal)
        assert_equal(res.unit, units.dimensionless_unscaled)

        with pytest.raises(UnitsError, match='log is only defined for '
                           'dimensionless quantities.'):
            res = np.log(qf4)

        res = np.log(qf5)
        assert_equal(res.nominal, np.log(1.0))
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

    def test_qfloat_np_multiply(self):
        qf1 = QFloat(2.0, 0.2, 'm')
        qf2 = QFloat(1.2, 0.1, 'm')
        qf3 = QFloat([1, 2, 4], [0.1, 0.2, 0.4], 'cm')

        res = np.multiply(qf1, 2)
        assert_equal(res.nominal, 4)
        assert_equal(res.std_dev, 0.4)
        assert_equal(res.unit, units.Unit('m'))

        res = np.multiply(qf1, qf2)
        assert_equal(res.nominal, 2.4)
        assert_almost_equal(res.std_dev, 0.3124099870362662)
        assert_equal(res.unit, units.Unit('m2'))

        res = np.multiply(qf1, qf3)
        assert_equal(res.nominal, [2, 4, 8])
        assert_almost_equal(res.std_dev, [0.28284271, 0.56568542, 1.13137085])
        assert_equal(res.unit, units.Unit('m*cm'))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.multiply(qf1, qf2, out=[])

    def test_qfloat_np_negative(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(-1.0, 0.1, 'm')
        qf3 = QFloat(-5.0, 0.1)
        qf4 = QFloat(6)
        qf5 = QFloat([1, -1, 2, -2])

        assert_equal(np.negative(qf1), QFloat(-1.0, 0.1, 'm'))
        assert_equal(np.negative(qf2), QFloat(1.0, 0.1, 'm'))
        assert_equal(np.negative(qf3), QFloat(5.0, 0.1))
        assert_equal(np.negative(qf4), QFloat(-6))
        assert_equal(np.negative(qf5), QFloat([-1, 1, -2, 2]))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.negative(qf1, out=[])

    def test_qfloat_np_positive(self):
        qf1 = QFloat(1.0, 0.1, 'm')
        qf2 = QFloat(-1.0, 0.1, 'm')
        qf3 = QFloat(-5.0, 0.1)
        qf4 = QFloat(6)
        qf5 = QFloat([1, -1, 2, -2])

        assert_equal(np.positive(qf1), QFloat(1.0, 0.1, 'm'))
        assert_equal(np.positive(qf2), QFloat(-1.0, 0.1, 'm'))
        assert_equal(np.positive(qf3), QFloat(-5.0, 0.1))
        assert_equal(np.positive(qf4), QFloat(6))
        assert_equal(np.positive(qf5), QFloat([1, -1, 2, -2]))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.positive(qf1, out=[])

    @pytest.mark.parametrize('func', [np.power, np.float_power])
    def test_qfloat_np_power(self, func):
        qf1 = QFloat(2.0, 0.1, 'm')
        qf2 = QFloat([2, 3, 4], [0.1, 0.2, 0.3], 'm')
        qf3 = QFloat(2.0, 0.1)
        qf4 = QFloat([2, 3, 4])

        res = func(qf1, 2)
        assert_equal(res.nominal, 4)
        assert_equal(res.std_dev, 0.4)
        assert_equal(res.unit, units.Unit('m2'))

        res = func(qf1, 1.5)
        assert_almost_equal(res.nominal, 2.8284271247461903)
        assert_almost_equal(res.std_dev, 0.2121320343559643)
        assert_equal(res.unit, units.Unit('m(3/2)'))

        res = func(qf2, 2)
        assert_equal(res.nominal, [4, 9, 16])
        assert_almost_equal(res.std_dev, [0.4, 1.2, 2.4])
        assert_equal(res.unit, units.Unit('m2'))

        res = func(qf2, 1.5)
        assert_almost_equal(res.nominal, [2.82842712, 5.19615242, 8])
        assert_almost_equal(res.std_dev, [0.21213203, 0.51961524, 0.9])
        assert_equal(res.unit, units.Unit('m(3/2)'))

        res = func(qf1, qf3)
        assert_equal(res.nominal, 4)
        assert_almost_equal(res.std_dev, 0.4866954717550927)
        assert_equal(res.unit, units.Unit('m2'))

        with pytest.raises(ValueError):
            func(qf1, qf4)

        with pytest.raises(ValueError):
            func(qf2, qf4)

        with pytest.raises(ValueError):
            func(qf4, qf1)

        with pytest.raises(NotImplementedError):
            # out argument should fail
            func(qf1, 2, out=[])

    @pytest.mark.parametrize('func', [np.mod, np.remainder])
    def test_qfloat_np_remainder(self, func):
        qf1 = QFloat(5.0, 0.1, 'm')
        qf2 = QFloat(3.5, 0.1, 'm')
        qf3 = QFloat(1.0, 0.1, 's')
        qf4 = QFloat([1, 2, 3])

        res = func(qf1, 2)
        assert_equal(res.nominal, 1)
        assert_equal(res.std_dev, 0.1)
        assert_equal(res.unit, units.Unit('m'))

        res = func(qf1, qf2)
        assert_equal(res.nominal, 1.5)
        assert_equal(res.std_dev, 0.14142135623730953)
        assert_equal(res.unit, units.Unit('m'))

        res = func(qf1, qf3)
        assert_equal(res.nominal, 0)
        assert_equal(res.std_dev, np.inf)
        assert_equal(res.unit, units.Unit('m'))

        res = func(qf1, qf4)
        assert_equal(res.nominal, [0, 1, 2])
        assert_equal(res.std_dev, [np.nan, 0.1, 0.1])
        assert_equal(res.unit, units.Unit('m'))

        res = func(qf4, 1.5)
        assert_equal(res.nominal, [1, 0.5, 0])
        assert_equal(res.std_dev, [0, 0, np.nan])
        assert_equal(res.unit, units.Unit(''))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            func(qf1, qf2, out=[])

    @pytest.mark.skip(reason="Not Implemented Yet")
    def test_qfloat_np_rint(self):
        raise NotImplementedError

    def test_qfloat_np_sign(self):
        assert_equal(np.sign(QFloat(1.5, 0.1, 'm')), 1)
        assert_equal(np.sign(QFloat(-1.5, 0.1, 'm')), -1)
        assert_equal(np.sign(QFloat(0.0, 0.1, 'm')), 0)

    def test_qfloat_np_signbit(self):
        assert_equal(np.sign(QFloat(1.5, 0.1, 'm')), 1)
        assert_equal(np.sign(QFloat(-1.5, 0.1, 'm')), -1)
        assert_equal(np.sign(QFloat(0.0, 0.1, 'm')), 0)
        assert_equal(np.sign(QFloat(-0.0, 0.1, 'm')), -0)

    def test_qfloat_np_sqrt(self):
        qf1 = QFloat(4, 0.1, 'm2')
        qf2 = QFloat([9, 100], [0.1, 0.1], 's2')

        res = np.sqrt(qf1)
        assert_equal(res.nominal, 2)
        assert_equal(res.std_dev, 0.025)
        assert_equal(res.unit, units.Unit('m'))

        res = np.sqrt(qf2)
        assert_equal(res.nominal, [3, 10])
        assert_almost_equal(res.std_dev, [0.01666667, 0.005])
        assert_equal(res.unit, units.Unit('s'))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.sqrt(qf1, out=[])

    def test_qfloat_np_square(self):
        qf1 = QFloat(2.0, 0.1, 'm')
        qf2 = QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'cm')

        res = np.square(qf1)
        assert_equal(res.nominal, 4)
        assert_almost_equal(res.std_dev, 0.28284271247461906)
        assert_equal(res.unit, units.Unit('m2'))

        res = np.square(qf2)
        assert_equal(res.nominal, [1, 4, 9])
        assert_almost_equal(res.std_dev, [0.14142136, 0.56568542, 1.27279221])
        assert_equal(res.unit, units.Unit('cm2'))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.square(qf1, out=[])

    def test_qfloat_np_subtract(self):
        qf1 = QFloat(2.0, 0.2, 'm')
        qf2 = QFloat(1.0, 0.1, 'm')
        qf3 = QFloat([1, 2, 3], [0.1, 0.2, 0.3], 'm')
        qf4 = QFloat(1.0, 0.1, 's')
        qf5 = QFloat(1.0)

        res = np.subtract(qf1, qf2)
        assert_equal(res.nominal, 1.0)
        assert_almost_equal(res.std_dev, 0.223606797749979)
        assert_equal(res.unit, units.Unit('m'))

        res = np.subtract(qf1, qf3)
        assert_equal(res.nominal, [1, 0, -1])
        assert_almost_equal(res.std_dev, [0.2236068, 0.28284271, 0.36055513])
        assert_equal(res.unit, units.Unit('m'))

        res = np.subtract(qf3, qf1)
        assert_equal(res.nominal, [-1, 0, 1])
        assert_almost_equal(res.std_dev, [0.2236068, 0.28284271, 0.36055513])
        assert_equal(res.unit, units.Unit('m'))

        with pytest.raises(UnitsError):
            np.subtract(qf1, qf4)

        with pytest.raises(UnitsError):
            np.subtract(qf1, qf5)

        with pytest.raises(UnitsError):
            np.subtract(qf1, 1.0)

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.subtract(qf1, qf2, out=[])

    def test_qfloat_np_trunc(self):
        assert_equal(np.trunc(QFloat(1.5, 0.1, 'm')), QFloat(1, 0.0, 'm'))
        assert_equal(np.trunc(QFloat(-1.5, 0.1, 'm')), QFloat(-1, 0.0, 'm'))


class TestQFloatNumpyUfuncTrigonometric:
    """Test the numpy trigonometric and inverse trigonometric functions."""

    # Both radians and deg2rad must work in the same way
    @pytest.mark.parametrize('func', [np.radians, np.deg2rad])
    def test_qfloat_np_radians(self, func):
        qf = QFloat(180, 0.1, 'degree')
        res = func(qf)
        assert_almost_equal(res.nominal, 3.141592653589793)
        assert_almost_equal(res.std_dev, 0.001745329251994)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat(-180, 0.1, 'degree')
        res = func(qf)
        assert_almost_equal(res.nominal, -3.141592653589793)
        assert_almost_equal(res.std_dev, 0.001745329251994)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([0, 30, 45, 60, 90], [0.1, 0.2, 0.3, 0.4, 0.5], 'degree')
        res = func(qf)
        assert_almost_equal(res.nominal, [0, 0.52359878, 0.78539816,
                                          1.04719755, 1.57079633])
        assert_almost_equal(res.std_dev, [0.00174533, 0.00349066, 0.00523599,
                                          0.00698132, 0.00872665])
        assert_equal(res.unit, units.Unit('rad'))

        # radian should no change
        qf = QFloat(1.0, 0.1, 'radian')
        res = func(qf)
        assert_equal(res.nominal, 1.0)
        assert_equal(res.std_dev, 0.1)
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ('m', None, 'm/s'):
            with pytest.raises(UnitsError):
                func(QFloat(1.0, 0.1, unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            func(qf, out=[])

    # Both degrees and rad2deg must work in the same way
    @pytest.mark.parametrize('func', [np.degrees, np.rad2deg])
    def test_qfloat_np_degrees(self, func):
        qf = QFloat(np.pi, 0.05, 'radian')
        res = func(qf)
        assert_almost_equal(res.nominal, 180.0)
        assert_almost_equal(res.std_dev, 2.8647889756541165)
        assert_equal(res.unit, units.Unit('deg'))

        qf = QFloat(-np.pi, 0.05, 'radian')
        res = func(qf)
        assert_almost_equal(res.nominal, -180.0)
        assert_almost_equal(res.std_dev, 2.8647889756541165)
        assert_equal(res.unit, units.Unit('deg'))

        qf = QFloat([np.pi, np.pi/2, np.pi/4, np.pi/6],
                    [0.01, 0.02, 0.03, 0.04], 'rad')
        res = func(qf)
        assert_almost_equal(res.nominal, [180.0, 90.0, 45.0, 30.0])
        assert_almost_equal(res.std_dev, [0.5729578, 1.14591559,
                                          1.71887339, 2.29183118])
        assert_equal(res.unit, units.Unit('deg'))

        # deg should no change
        qf = QFloat(1.0, 0.1, 'deg')
        res = func(qf)
        assert_equal(res.nominal, 1.0)
        assert_equal(res.std_dev, 0.1)
        assert_equal(res.unit, units.Unit('deg'))

        # Invalid units
        for unit in ('m', None, 'm/s'):
            with pytest.raises(UnitsError):
                func(QFloat(1.0, 0.1, unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            func(qf, out=[])

    def test_qfloat_np_sin(self):
        qf = QFloat(np.pi, 0.05, 'radian')
        res = np.sin(qf)
        assert_almost_equal(res.nominal, 0.0)
        assert_almost_equal(res.std_dev, 0.05)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(90, 0.05, 'deg')
        res = np.sin(qf)
        assert_almost_equal(res.nominal, 1.0)
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat([30, 45, 60], [0.1, 0.2, 0.3], 'deg')
        res = np.sin(qf)
        assert_almost_equal(res.nominal, [0.5, 0.70710678, 0.8660254])
        assert_almost_equal(res.std_dev, [0.0015115, 0.00246827, 0.00261799])
        assert_equal(res.unit, units.dimensionless_unscaled)

        for unit in ['m', 'm/s', None]:
            with pytest.raises(UnitsError):
                np.sin(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.sin(qf, out=[])

    def test_qfloat_np_cos(self):
        qf = QFloat(180, 0.05, 'deg')
        res = np.cos(qf)
        assert_almost_equal(res.nominal, -1.0)
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(np.pi/2, 0.05, 'rad')
        res = np.cos(qf)
        assert_almost_equal(res.nominal, 0.0)
        assert_almost_equal(res.std_dev, 0.05)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat([30, 45, 60], [0.1, 0.2, 0.3], 'deg')
        res = np.cos(qf)
        assert_almost_equal(res.nominal, [0.8660254, 0.70710678, 0.5])
        assert_almost_equal(res.std_dev, [0.00087266, 0.00246827, 0.0045345])
        assert_equal(res.unit, units.dimensionless_unscaled)

        for unit in ['m', 'm/s', None]:
            with pytest.raises(UnitsError):
                np.cos(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.cos(qf, out=[])

    def test_qfloat_np_tan(self):
        qf = QFloat(45, 0.05, 'deg')
        res = np.tan(qf)
        assert_almost_equal(res.nominal, 1.0)
        assert_almost_equal(res.std_dev, 0.0017453292519943294)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(np.pi/4, 0.05, 'rad')
        res = np.tan(qf)
        assert_almost_equal(res.nominal, 1.0)
        assert_almost_equal(res.std_dev, 0.1)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat([0, 30, 60], [0.1, 0.2, 0.3], 'deg')
        res = np.tan(qf)
        assert_almost_equal(res.nominal, [0, 0.57735027, 1.73205081])
        assert_almost_equal(res.std_dev, [0.00174533, 0.00465421, 0.02094395])
        assert_equal(res.unit, units.dimensionless_unscaled)

        for unit in ['m', 'm/s', None]:
            with pytest.raises(UnitsError):
                np.tan(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.tan(qf, out=[])

    def test_qfloat_np_sinh(self):
        qf = QFloat(0, 0.05, 'radian')
        res = np.sinh(qf)
        assert_almost_equal(res.nominal, 0.0)
        assert_almost_equal(res.std_dev, 0.05)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(np.pi, 0.05, 'radian')
        res = np.sinh(qf)
        assert_almost_equal(res.nominal, 11.548739357257748)
        assert_almost_equal(res.std_dev, 0.5795976637760759)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(90, 0.05, 'deg')
        res = np.sinh(qf)
        assert_almost_equal(res.nominal, 2.3012989023072947)
        assert_almost_equal(res.std_dev, 0.002189671298638268)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat([30, 45, 60], [0.1, 0.2, 0.3], 'deg')
        res = np.sinh(qf)
        assert_almost_equal(res.nominal, [0.5478535, 0.86867096, 1.24936705])
        assert_almost_equal(res.std_dev, [0.0019901, 0.0046238, 0.0083791])
        assert_equal(res.unit, units.dimensionless_unscaled)

        for unit in ['m', 'm/s', None]:
            with pytest.raises(UnitsError):
                np.sinh(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.sinh(qf, out=[])

    def test_qfloat_np_cosh(self):
        qf = QFloat(0, 0.05, 'radian')
        res = np.cosh(qf)
        assert_almost_equal(res.nominal, 1.0)
        assert_almost_equal(res.std_dev, 0.0)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(np.pi, 0.05, 'radian')
        res = np.cosh(qf)
        assert_almost_equal(res.nominal, 11.591953275521519)
        assert_almost_equal(res.std_dev, 0.5774369678628875)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(90, 0.05, 'deg')
        res = np.cosh(qf)
        assert_almost_equal(res.nominal, 2.5091784786580567)
        assert_almost_equal(res.std_dev, 0.0020082621458896)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat([30, 45, 60], [0.1, 0.2, 0.3], 'deg')
        res = np.cosh(qf)
        assert_almost_equal(res.nominal, [1.14023832, 1.32460909, 1.60028686])
        assert_almost_equal(res.std_dev, [0.00095618, 0.00303223, 0.00654167])
        assert_equal(res.unit, units.dimensionless_unscaled)

        for unit in ['m', 'm/s', None]:
            with pytest.raises(UnitsError):
                np.cosh(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.cosh(qf, out=[])

    def test_qfloat_np_tanh(self):
        qf = QFloat(0, 0.05, 'radian')
        res = np.tanh(qf)
        assert_almost_equal(res.nominal, 0.0)
        assert_almost_equal(res.std_dev, 0.05)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(np.pi, 0.05, 'radian')
        res = np.tanh(qf)
        assert_almost_equal(res.nominal, 0.99627207622075)
        assert_almost_equal(res.std_dev, 0.00037209750714)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat(90, 0.05, 'deg')
        res = np.tanh(qf)
        assert_almost_equal(res.nominal, 0.9171523356672744)
        assert_almost_equal(res.std_dev, 0.0001386067128590)
        assert_equal(res.unit, units.dimensionless_unscaled)

        qf = QFloat([30, 45, 60], [0.1, 0.2, 0.3], 'deg')
        res = np.tanh(qf)
        assert_almost_equal(res.nominal, [0.48047278, 0.6557942, 0.78071444])
        assert_almost_equal(res.std_dev, [0.00134241, 0.00198944, 0.00204457])
        assert_equal(res.unit, units.dimensionless_unscaled)

        for unit in ['m', 'm/s', None]:
            with pytest.raises(UnitsError):
                np.tanh(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.tanh(qf, out=[])

    def test_qfloat_np_arcsin(self):
        qf = QFloat(np.sqrt(2)/2, 0.01)
        res = np.arcsin(qf)
        assert_almost_equal(res.nominal, 0.7853981633974484)
        assert_almost_equal(res.std_dev, 0.0141421356237309)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([0, 0.5, 1], [0.01, 0.2, 0.3])
        res = np.arcsin(qf)
        assert_almost_equal(res.nominal, [0, 0.52359878, 1.57079633])
        assert_almost_equal(res.std_dev, [0.01, 0.23094011, np.inf])
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ['m', 'm/s', 'rad', 'deg']:
            with pytest.raises(UnitsError):
                np.arcsin(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.arcsin(qf, out=[])

    def test_qfloat_np_arccos(self):
        qf = QFloat(np.sqrt(2)/2, 0.01)
        res = np.arccos(qf)
        assert_almost_equal(res.nominal, 0.7853981633974484)
        assert_almost_equal(res.std_dev, 0.0141421356237309)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([0, 0.5, 1], [0.01, 0.2, 0.3])
        res = np.arccos(qf)
        assert_almost_equal(res.nominal, [1.57079633, 1.04719755, 0])
        assert_almost_equal(res.std_dev, [0.01, 0.23094011, np.inf])
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ['m', 'm/s', 'rad', 'deg']:
            with pytest.raises(UnitsError):
                np.arccos(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.arccos(qf, out=[])

    def test_qfloat_np_arctan(self):
        qf = QFloat(1.0, 0.01)
        res = np.arctan(qf)
        assert_almost_equal(res.nominal, 0.7853981633974484)
        assert_almost_equal(res.std_dev, 0.005)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([0, 0.5, 1], [0.01, 0.2, 0.3])
        res = np.arctan(qf)
        assert_almost_equal(res.nominal, [0, 0.4636476, 0.7853982])
        assert_almost_equal(res.std_dev, [0.01, 0.16, 0.15])
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ['m', 'm/s', 'rad', 'deg']:
            with pytest.raises(UnitsError):
                np.arctan(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.arctan(qf, out=[])

    def test_qfloat_np_arcsinh(self):
        qf = QFloat(0.0, 0.01)
        res = np.arcsinh(qf)
        assert_almost_equal(res.nominal, 0.0)
        assert_almost_equal(res.std_dev, 0.01)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([0.5, 1.0, 10], [0.01, 0.2, 0.3])
        res = np.arcsinh(qf)
        assert_almost_equal(res.nominal, [0.4812118, 0.8813736, 2.998223])
        assert_almost_equal(res.std_dev, [0.0089443, 0.1414214, 0.0298511])
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ['m', 'm/s', 'rad', 'deg']:
            with pytest.raises(UnitsError):
                np.arcsinh(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.arcsinh(qf, out=[])

    def test_qfloat_np_arccosh(self):
        qf = QFloat(1.0, 0.01)
        res = np.arccosh(qf)
        assert_almost_equal(res.nominal, 0.0)
        # assert_almost_equal(res.std_dev, np.inf)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([1.5, 5.0, 10], [0.01, 0.2, 0.3])
        res = np.arccosh(qf)
        assert_almost_equal(res.nominal, [0.9624237, 2.2924317, 2.9932228])
        assert_almost_equal(res.std_dev, [0.0089443, 0.0408248, 0.0301511])
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ['m', 'm/s', 'rad', 'deg']:
            with pytest.raises(UnitsError):
                np.arccosh(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.arccosh(qf, out=[])

    def test_qfloat_np_arctanh(self):
        qf = QFloat(0.0, 0.01)
        res = np.arctanh(qf)
        assert_almost_equal(res.nominal, 0.0)
        assert_almost_equal(res.std_dev, 0.01)
        assert_equal(res.unit, units.Unit('rad'))

        qf = QFloat([0.1, 0.5, 1.0], [0.01, 0.2, 0.3])
        res = np.arctanh(qf)
        assert_almost_equal(res.nominal, [0.1003353, 0.5493061, np.inf])
        assert_almost_equal(res.std_dev, [0.010101, 0.2666667, np.inf])
        assert_equal(res.unit, units.Unit('rad'))

        # Invalid units
        for unit in ['m', 'm/s', 'rad', 'deg']:
            with pytest.raises(UnitsError):
                np.arctanh(QFloat(1.0, unit=unit))

        with pytest.raises(NotImplementedError):
            # out argument should fail
            np.arctanh(qf, out=[])

    def test_qfloat_np_arctan2(self):
        qf1 = QFloat(1.0, 0.01)
        qf2 = QFloat(0.0, 0.01)
        res = np.arctan2(qf1, qf2)
        assert_almost_equal(res.nominal, 1.57079633)
        assert_almost_equal(res.std_dev, 0.01)
        assert_equal(res.unit, units.Unit('rad'))

        qf1 = QFloat([0.5, 1.0, 10], [0.01, 0.2, 0.3])
        qf2 = QFloat([0.1, 0.5, 1.0], [0.01, 0.2, 0.3])
        res = np.arctan2(qf1, qf2)
        assert_almost_equal(res.nominal, [1.373401, 1.107149, 1.471128])
        assert_almost_equal(res.std_dev, [0.019612, 0.178885, 0.029851])
        assert_equal(res.unit, units.Unit('rad'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import logging


__all__ = ['logger', 'ListHandler', 'log_to_list', 'resolve_level_string']


logging.basicConfig(format='%(asctime)-15s %(name)s - %(levelname)s -'
                           ' %(message)s  [%(module)s]')
logger = logging.getLogger('astropop')
logger.setLevel('INFO')


class ListHandler(logging.Handler):
    """Logging handler to save messages in a list. No thread safe!

    Parameters
    ----------
    log_list : list or None, optional
        List to store the logs. If None, a new list will be created and
        can be accesses with `ListHandler.log_list`
    full_record : bool, optional
        Store the full Python log record instead of just the message.
    """
    def __init__(self, log_list=None, full_record=False):
        logging.Handler.__init__(self)
        if log_list is None:
            log_list = []
        self._log_list = log_list
        self._full_record = full_record

    @property
    def log_list(self):
        """List where records are being stored."""
        return self._log_list

    @property
    def full_record(self):
        """True if full records are being stored in list. False if only
        messages."""
        return self._full_record

    def emit(self, record):
        """Append the log record to the list."""
        if self.full_record:
            self.log_list.append(record)
        else:
            self.log_list.append(record.getMessage())


def log_to_list(logger, log_list=None, full_record=False):
    """Add a ListHandler and a log_list to a Logger.

    Parameters
    ----------
    logger : `logging.Logger`
        `Logger` instance where handler will be added.
    log_list : list or None, optional
        List to store log records.
    full_record : bool, optional
        Store full log records instead of just the message.

    Returns
    -------
        `ListHandler` created during the process.
    """
    lh = ListHandler(log_list, full_record)
    logger.addHandler(lh)
    return lh


def resolve_level_string(value):
    """Resolve the log level of a string."""
    try:
        # Handle already int values
        value = int(value)
        return value
    except ValueError:
        return getattr(logging, value)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utils to handle FITS files."""

from astropy.io import fits
from .logger import logger


__all__ = ['imhdus', 'check_header_keys']


_fits_extensions = ['.fits', '.fts', '.fit', '.fz']
_fits_compression = ['.gz', '.zip', '.bz2', '.Z']
_fits_extensions_with_compress = _fits_extensions.copy()

for k in _fits_compression:
    _fits_extensions_with_compress.extend(
        [i+k for i in _fits_extensions_with_compress]
    )


imhdus = (fits.ImageHDU, fits.PrimaryHDU, fits.CompImageHDU,
          fits.StreamingHDU)


class IncompatibleHeadersError(ValueError):
    """When 2 image header are not compatible."""


def check_header_keys(image1, image2, keywords=None):
    """Compare header keys from 2 images to check if the have equal values."""
    keywords = keywords or []
    header1 = {}
    header2 = {}

    # Compatibility with fits HDU and FrameData
    if hasattr(image1, 'header'):
        header1 = image1.header
    elif hasattr(image1, 'meta'):
        header1 = image1.meta

    if hasattr(image2, 'header'):
        header2 = image2.header
    elif hasattr(image2, 'meta'):
        header2 = image2.meta

    for i in keywords:
        if i in header1 and i in header2:
            v1 = header1[i]
            v2 = header2[i]
            if v1 != v2:
                raise IncompatibleHeadersError(f'Keyword `{i}` have different '
                                               'values for images 1 and 2:'
                                               '`{v1}`  `{v2}`')
        elif i in header1 or i in header2:
            raise IncompatibleHeadersError("Headers have inconsisten presence "
                                           f"of {i} Keyword")
        else:
            logger.debug("The images do not have the %s keyword", i)
    return True



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Query and match objects in Vizier catalogs."""

from functools import partial
import numpy as np
from multiprocessing.pool import Pool
from astroquery.simbad import Simbad, SimbadClass
from astropy.coordinates import SkyCoord
from astropy import units as u

from .base_catalog import _BasePhotometryCatalog
from ._online_tools import _timeout_retry, \
                           MAX_PARALLEL_QUERY, \
                           astroquery_radius, astroquery_skycoord
from ..py_utils import string_fix


__all__ = ['simbad_query_id', 'SimbadCatalog', 'SimbadCatalogClass']


def simbad_query_id(ra, dec, limit_angle, name_order=None,
                    simbad=None):
    """Query name ids for a star in Simbad.

    Parameters
    ----------
    ra, dec : `float`
        RA and DEC decimal degrees coordinates to query.
    limit_angle : string, float, `~astropy.coordinates.Angle`
        Maximum radius for search.
    name_order : `list`, optional
        Order of priority of name prefixes to query.
        Default: None
    simbad : `~astroquery.simbad.Simbad`, optional
        `~astroquery.simbad.Simbad` to be used in query.

    Returns
    -------
    `str`
        The ID of the object.
    """
    if name_order is None:
        name_order = ['MAIN_ID', 'NAME', 'HD', 'HR', 'HYP', 'TYC', 'AAVSO']

    if simbad is not None:
        s = simbad
    else:
        s = Simbad()

    def _strip_spaces(name):
        name = name.strip('NAME')
        name = name.strip('* ')
        # remove excessive spaces
        while '  ' in name:
            name = name.replace('  ', ' ')
        return name.strip(' ')

    q = _timeout_retry(s.query_region, SkyCoord(ra, dec,
                                                unit=(u.degree, u.degree)),
                       radius=limit_angle)

    if q is not None:
        name = string_fix(q['MAIN_ID'][0])
        ids = _timeout_retry(s.query_objectids, name)['ID']
        for i in name_order:
            if i == 'MAIN_ID':
                return _strip_spaces(name)
            for k in ids:
                if i+' ' in k:
                    return _strip_spaces(k)
    return None


class SimbadCatalogClass(_BasePhotometryCatalog):
    """Base class to handle with Simbad."""

    id_key = 'MAIN_ID'
    ra_key = 'RA'
    dec_key = 'DEC'
    flux_key = 'FLUX_{band}'
    flux_error_key = 'FLUX_ERROR_{band}'
    flux_unit_key = 'FLUX_UNIT_{band}'
    flux_bibcode_key = 'FLUX_BIBCODE_{band}'
    type = 'online'
    prepend_id_key = False
    available_filters = ["U", "B", "V", "R", "I", "J", "H", "K", "u", "g", "r",
                         "i", "z"]
    _last_query_info = None
    _last_query_table = None
    _simbad = None

    @property
    def simbad(self):
        """Query operator instance."""
        if self._simbad is None:
            self._simbad = Simbad()
        self._simbad.ROW_LIMIT = 0
        return self._simbad()

    @simbad.setter
    def simbad(self, value):
        if not isinstance(value, SimbadClass):
            raise ValueError(f"{value} is not a SimbadClass instance.")
        self._simbad = value

    def get_simbad(self, band=None):
        """Get a copy of the simbad querier with optional band fields."""
        s = self.simbad
        if band is not None:
            s.add_votable_fields(f'fluxdata({band})')
        return s

    def _flux_keys(self, band):
        flux_key = self.flux_key.format(band=band)
        flux_error_key = self.flux_error_key.format(band=band)
        flux_bibcode_key = self.flux_bibcode_key.format(band=band)
        flux_unit_key = self.flux_unit_key.format(band=band)
        return flux_key, flux_error_key, flux_unit_key, flux_bibcode_key

    def query_object(self, center, band=None, **kwargs):
        """Query a single object in the catalog."""
        s = self.get_simbad(band)
        if not isinstance(center, str):
            raise ValueError("Simbad query_object only accept object name. "
                             "Try query_region instead.")
        return self._query(s.query_object, center, **kwargs)

    def query_region(self, center, radius, band=None, **kwargs):
        """Query all objects in a region."""
        s = self.get_simbad(band)
        center = astroquery_skycoord(center)
        radius = astroquery_radius(radius)
        return self._query(s.query_region, center, radius=radius, **kwargs)

    def _id_resolve(self, idn):
        if self.prepend_id_key:
            idn = [f"{self.id_key} {i}" for i in idn]
            idn = np.array(idn)
        return np.array([string_fix(i) for i in idn])

    def filter_flux(self, band, query=None):
        """Filter the flux data of a query."""
        self.check_filter(band)
        flux_key, error_key, unit_key, bibcode_key = self._flux_keys(band)
        if query is None:
            query = self._last_query_table

        if flux_key not in query.colnames:
            raise KeyError(f'Simbad query must be performed with band {band}'
                           ' for flux data.')

        flux = np.array(query[flux_key].data)
        flux_error = np.array(query[error_key].data)
        bibcode = np.array(query[bibcode_key].data)
        unit = np.array(query[unit_key].data)

        return flux, flux_error, unit, bibcode

    def match_objects(self, ra, dec, band=None, limit_angle='2 arcsec'):
        """Match objects from RA DEC list with this catalog."""
        flux_keys = ['flux', 'flux_error', 'flux_unit', 'flux_bibcode']
        table_props = [('id', ''), ('ra', np.nan), ('dec', np.nan),
                       ('flux', np.nan), ('flux_error', np.nan),
                       ('flux_unit', ''), 'flux_bibcode', '']
        res = self._match_objects(ra, dec, band, limit_angle,
                                  flux_keys, table_props)

        return res

    @staticmethod
    def match_object_ids(ra, dec, limit_angle='2 arcsec', name_order=None):
        """Get the id from Simbad for every object in a RA, Dec list."""
        # Perform it in parallel to handle the online query overhead
        func = partial(simbad_query_id, name_order=name_order,
                       limit_angle=limit_angle)
        p = Pool(MAX_PARALLEL_QUERY)
        results = p.starmap(func, list(zip(ra, dec)))
        return results


SimbadCatalog = SimbadCatalogClass()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import numpy as np
import pytest

from astropop.image.imarith import imarith
from astropop.framedata import FrameData
from astropop.testing import *


pars = pytest.mark.parametrize('op,vs', [('+', {'f1': {'v': 30, 'u': 0},
                                                'f2': {'v': 10, 'u': 0},
                                                'r': {'v': 40, 'u': 0}}),
                                         ('+', {'f1': {'v': 30, 'u': 3},
                                                'f2': {'v': 10, 'u': 4},
                                                'r': {'v': 40, 'u': 5}}),
                                         ('-', {'f1': {'v': 30, 'u': 0},
                                                'f2': {'v': 10, 'u': 0},
                                                'r': {'v': 20, 'u': 0}}),
                                         ('-', {'f1': {'v': 30, 'u': 3},
                                                'f2': {'v': 10, 'u': 4},
                                                'r': {'v': 20, 'u': 5}}),
                                         ('*', {'f1': {'v': 5, 'u': 0},
                                                'f2': {'v': 6, 'u': 0},
                                                'r': {'v': 30, 'u': 0}}),
                                         ('*', {'f1': {'v': 5, 'u': 0.3},
                                                'f2': {'v': 6, 'u': 0.4},
                                                'r': {'v': 30,
                                                      'u': 2.690725}}),
                                         ('/', {'f1': {'v': 10, 'u': 0},
                                                'f2': {'v': 3, 'u': 0},
                                                'r': {'v': 3.33333333,
                                                      'u': 0}}),
                                         ('/', {'f1': {'v': 10, 'u': 1},
                                                'f2': {'v': 3, 'u': 0.3},
                                                'r': {'v': 3.33333333,
                                                      'u': 0.47140452}}),
                                         ('//', {'f1': {'v': 10, 'u': 0},
                                                 'f2': {'v': 3, 'u': 0},
                                                 'r': {'v': 3.000000,
                                                       'u': 0}}),
                                         ('//', {'f1': {'v': 10, 'u': 1},
                                                 'f2': {'v': 3, 'u': 0.3},
                                                 'r': {'v': 3.000000,
                                                       'u': 0.000000}})])


@pytest.mark.parametrize('handle_mask', [True, False])
@pytest.mark.parametrize('inplace', [True, False])
@pars
def test_imarith_ops_frames(op, vs, inplace, handle_mask):
    def gen_frame(v):
        # Gen frames with {'v', 'u'} dict
        shape = (10, 10)
        if v['u'] is None:
            frame = FrameData(np.ones(shape, dtype='f8'), unit='adu')
        else:
            frame = FrameData(np.ones(shape, dtype='f8'), unit='adu',
                              uncertainty=v['u'])
        frame.data[:] = v['v']
        return frame

    frame1 = gen_frame(vs['f1'])
    frame2 = gen_frame(vs['f2'])
    exp_res = gen_frame(vs['r'])
    if handle_mask:
        mask1 = np.zeros((10, 10))
        mask2 = np.zeros((10, 10))
        mask1[5, 5] = 1
        mask2[3, 3] = 1
        exp_mask = np.zeros((10, 10))
        exp_mask[5, 5] = 1
        exp_mask[3, 3] = 1
        frame1.mask = mask1
        frame2.mask = mask2
        exp_res.mask = exp_mask

    res = imarith(frame1, frame2, op, inplace=inplace,
                  join_masks=handle_mask)

    assert_almost_equal(res.data, exp_res.data)
    assert_almost_equal(res.uncertainty, exp_res.uncertainty)
    if handle_mask:
        assert_equal(res.mask, exp_res.mask)

    if inplace:
        assert_is(res, frame1)
    else:
        assert_is_not(res, frame1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np

from astropy import units as u
from astropy.wcs import WCS
from astropop.image.processing import cosmics_lacosmic, \
                                      gain_correct, \
                                      subtract_bias, \
                                      subtract_dark, \
                                      flat_correct, \
                                      trim_image
from astropop.framedata import FrameData
from astropop.testing import *


class Test_Processing_Cosmics():
    @pytest.mark.parametrize('inplace', [True, False])
    def test_cosmics_lacosmic(self, inplace):
        image = FrameData(np.ones((20, 20))*3, unit=u.adu,
                          mask=np.zeros((20, 20)))
        # Add cosmics
        image.data[10, 10] = 35000
        image.data[10, 11] = 35000
        image.data[11, 10] = 35000
        image.mask[15, 18] = 1

        expect_mask = np.zeros((20, 20))
        # currently we are not updating the mask
        # expect_mask[10, 10] = 1
        # expect_mask[10, 11] = 1
        # expect_mask[11, 10] = 1
        expect_mask[15, 18] = 1

        # Run the cosmics removal
        res = cosmics_lacosmic(image, inplace=inplace)
        assert_equal(res.data, np.ones((20, 20))*3)
        assert_equal(res.mask, expect_mask)
        assert_equal(res.meta['astropop lacosmic'], True)

        if inplace:
            assert_is(res, image)
        else:
            assert_is_not(res, image)


class Test_Processing_Gain():
    @pytest.mark.parametrize('inplace', [True, False])
    def test_simple_gain_correct(self, inplace):
        image = FrameData(np.ones((20, 20))*3, unit=u.adu,
                          uncertainty=5, mask=np.zeros((20, 20)))
        gain = 1.5*u.Unit('electron')/u.adu
        res = gain_correct(image, gain, inplace=inplace)

        assert_equal(res.data, np.ones((20, 20))*3*1.5)
        assert_equal(res.uncertainty, np.ones((20, 20))*5*1.5)
        assert_equal(res.unit, u.Unit('electron'))

        assert_equal(res.meta['astropop gain_corrected'], True)
        assert_equal(res.meta['astropop gain_corrected_value'], 1.5)
        assert_equal(res.meta['astropop gain_corrected_unit'],
                     'electron / adu')

        if inplace:
            assert_is(res, image)
        else:
            assert_is_not(res, image)


class Test_Processing_Flat():
    @pytest.mark.parametrize('inplace', [True, False])
    def test_simple_flat(self, inplace):
        expect = np.ones((20, 20))*3
        expect[0:5, 0:5] = 3/0.5

        # Checking flat division:
        frame1 = FrameData(np.ones((20, 20))*3, unit=u.adu)

        master_flat_dimless = FrameData(np.ones((20, 20)), unit=None)
        master_flat_dimless.data[0:5, 0:5] = 0.5

        res1 = flat_correct(frame1, master_flat_dimless, inplace=inplace)

        assert_is_instance(res1, FrameData)
        assert_equal(res1.data, expect)
        assert_equal(res1.header['astropop flat_corrected'], True)

        if inplace:
            assert_is(res1.data, frame1.data)
        else:
            assert_is_not(res1.data, frame1.data)


class Test_Processing_Bias():
    @pytest.mark.parametrize('inplace', [True, False])
    def test_simple_bias(self, inplace):
        expected = np.ones((20, 20))*2
        expected[0:5, 0:5] = 2.5

        frame4bias = FrameData(np.ones((20, 20))*3, unit='adu')

        master_bias = FrameData(np.ones((20, 20)), unit='adu')
        master_bias.data[0:5, 0:5] = 0.5

        res4 = subtract_bias(frame4bias, master_bias, inplace=inplace)

        assert_is_instance(res4, FrameData)
        assert_equal(res4.data, expected)
        assert_equal(res4.header['astropop bias_corrected'], True)

        if inplace:
            assert_is(res4.data, frame4bias.data)
        else:
            assert_is_not(res4.data, frame4bias.data)


@pytest.mark.skip
class Test_Processing_Dark():
    pass


class Test_Processing_TrimImage():
    @pytest.mark.parametrize('inplace', [True, False])
    def test_simple_trim(self, inplace):
        shape = (512, 1024)
        arr = np.random.uniform(20, 30, shape)
        mask = np.zeros(shape)
        mask[0:10, 0:40] = 1
        image = FrameData(arr, unit=u.adu, uncertainty=np.sqrt(arr),
                          mask=mask)

        # Trim the image
        xslice = slice(5, 25)
        yslice = slice(34, 46)
        res = trim_image(image, xslice, yslice, inplace=inplace)

        section = (yslice, xslice)

        assert_almost_equal(res.data, arr[section])
        assert_almost_equal(res.uncertainty, np.sqrt(arr)[section])
        assert_almost_equal(res.mask, mask[section])
        assert_equal(res.header['astropop trimmed_section'], '5:25,34:46')

        if inplace:
            assert_is(res, image)
        else:
            assert_is_not(res, image)

    @pytest.mark.parametrize('inplace', [True, False])
    def test_wcs_trim(self, inplace):
        shape = (512, 1024)
        arr = np.random.uniform(20, 30, shape)
        mask = np.zeros(shape)
        mask[0:10, 0:40] = 1
        wcs = WCS(naxis=2)
        wcs.wcs.crpix = [100, 100]
        wcs.wcs.cdelt = [1, 1]
        wcs.wcs.crval = [0, 0]
        wcs.wcs.ctype = ['RA---TAN', 'DEC--TAN']
        wcs.wcs.pc = [[1, 0], [0, 1]]

        image = FrameData(arr, unit=u.adu, uncertainty=np.sqrt(arr),
                          mask=mask, wcs=wcs)

        # Trim the image
        xslice = slice(5, 25)
        yslice = slice(34, 46)
        res = trim_image(image, xslice, yslice, inplace=inplace)

        section = (yslice, xslice)

        assert_almost_equal(res.data, arr[section])
        assert_almost_equal(res.uncertainty, np.sqrt(arr)[section])
        assert_almost_equal(res.mask, mask[section])
        assert_equal(res.header['astropop trimmed_section'], '5:25,34:46')
        assert_equal(res.wcs.wcs.crpix, [100-5, 100-34])
        assert_equal(res.wcs.wcs.cdelt, [1, 1])
        assert_equal(res.wcs.wcs.crval, [0, 0])
        assert_equal(res.wcs.wcs.ctype, ['RA---TAN', 'DEC--TAN'])
        assert_equal(res.wcs.wcs.pc, [[1, 0], [0, 1]])

        if inplace:
            assert_is(res, image)
        else:
            assert_is_not(res, image)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Collection of testing helpers.

Notes
-----
- Based on pytest_check. However, we are removing the dependency.
- Lots of functions imported directly from numpy.testing
"""

from os.path import exists
from numpy import testing as npt
from .py_utils import check_iterable


__all__ = ['assert_equal', 'assert_not_equal', 'assert_almost_equal',
           'assert_true', 'assert_false',
           'assert_is', 'assert_is_not',
           'assert_is_none', 'assert_is_not_none',
           'assert_in', 'assert_not_in',
           'assert_is_instance', 'assert_is_not_instance',
           'assert_greater', 'assert_greater_equal',
           'assert_less', 'assert_less_equal',
           'assert_raises', 'assert_raises_regex',
           'assert_warns', 'assert_not_warnings',
           'assert_path_exists', 'assert_path_not_exists']


def func_wrapper(func):
    """Warp the function, if needed. Now, its dummy."""
    # The idea here is to, in future, continue the tests and log all
    # failures at once. Like pytest_check.
    return func


# --------------------------------------------------------------------
# Testing Helpers
# --------------------------------------------------------------------

assert_raises = func_wrapper(npt.assert_raises)
assert_raises_regex = func_wrapper(npt.assert_raises_regex)
assert_warns = func_wrapper(npt.assert_warns)
assert_not_warnings = func_wrapper(npt.assert_no_warnings)


@func_wrapper
def assert_true(a, msg=None):
    """Raise assertion error if a is not True."""
    if not a:
        msg = msg or f"{a} is not True."
        raise AssertionError(msg)


@func_wrapper
def assert_false(a, msg=None):
    """Raise assertion error if condition is not False."""
    if a:
        msg = msg or f"{a} is not False."
        raise AssertionError(msg)


@func_wrapper
def assert_equal(a, b, msg=None):
    """Check if two objects are equal. Arrays supported."""
    if check_iterable(a) or check_iterable(b):
        if msg is None:
            msg = ''
        npt.assert_array_equal(a, b, err_msg=msg, verbose=True)
    else:
        if not a == b:
            msg = msg or f"{a} is not equal to {b}."
            raise AssertionError(msg)


@func_wrapper
def assert_almost_equal(a, b, decimal=6, msg=None):
    """Check if two objects are almost equal. Arrays supported."""
    if check_iterable(a) or check_iterable(b):
        if msg is None:
            msg = ''
        npt.assert_array_almost_equal(a, b, decimal=decimal,
                                      err_msg=msg, verbose=True)
    else:
        limit = 1.5 * 10**(-decimal)
        if not abs(a-b) < limit:
            msg = msg or f"{a} is not almost equal {b} within {limit}."
            raise AssertionError(msg)


@func_wrapper
def assert_not_equal(a, b, msg=None):
    """Raise assertion error if values are equal."""
    try:
        npt.assert_equal(a, b)
    except AssertionError:
        pass
    else:
        msg = msg or f"{a} is equal to {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_is(a, b, msg=None):
    """Raise assertion error if `a is b`."""
    if a is not b:
        msg = msg or f"{a} is not {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_is_not(a, b, msg=None):
    """Raise assertion error if `a is not b`."""
    if a is b:
        msg = msg or f"{a} is {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_is_none(a, msg=None):
    """Raise assertion error if `a is not None`."""
    if a is not None:
        msg = msg or f"{a} is not None."
        raise AssertionError(msg)


@func_wrapper
def assert_is_not_none(a, msg=None):
    """Raise assertion error if `a is None`."""
    if a is None:
        msg = msg or f"{a} is None."
        raise AssertionError(msg)


@func_wrapper
def assert_in(a, b, msg=None):
    """Raise assertion error if `a not in b`."""
    if a not in b:
        msg = msg or f"{a} is not in {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_not_in(a, b, msg=None):
    """Raise assertion error if `a in b`."""
    if a in b:
        msg = msg or f"{a} is in {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_is_instance(a, b, msg=None):
    """Raise assertion error if not `isinstance(a, b)`."""
    if not isinstance(a, b):
        msg = msg or f"{a} is not instance of {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_is_not_instance(a, b, msg=None):
    """Raise assertion error if `isinstance(a, b)`."""
    if isinstance(a, b):
        msg = msg or f"{a} is instance of {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_greater(a, b, msg=None):
    """Raise assertion error if a is not greater then b."""
    if not a > b:
        msg = msg or f"{a} is not greater than {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_greater_equal(a, b, msg=None):
    """Raise assertion error if a is not greater or equal then b."""
    if not a >= b:
        msg = msg or f"{a} is not greater or equal than {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_less(a, b, msg=None):
    """Raise assertion error if a is not less b."""
    if not a < b:
        msg = msg or f"{a} is not less than {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_less_equal(a, b, msg=None):
    """Raise assertion error if a is not less or equal then b."""
    if not a <= b:
        msg = msg or f"{a} is not less or equal than {b}."
        raise AssertionError(msg)


@func_wrapper
def assert_path_exists(a, msg=None):
    """Raise assertion error if path a do not exists."""
    if not exists(a):
        msg = msg or f"{a} do not exists."
        raise AssertionError(msg)


@func_wrapper
def assert_path_not_exists(a, msg=None):
    """Raise assertion error if path a do exists."""
    if exists(a):
        msg = msg or f"{a} do exists."
        raise AssertionError(msg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np

from astropop.photometry import (background, sepfind, daofind, starfind,
                                 calc_fwhm, recenter_sources)
from astropop.photometry.detection import gen_filter_kernel, DAOFind, \
                                          _sep_fix_byte_order
from astropop.framedata import MemMapArray
from astropop.math.moffat import moffat_2d
from astropop.math.gaussian import gaussian_2d
from astropop.math.array import trim_array
from astropy.utils import NumpyRNGContext
from astropy.stats import gaussian_fwhm_to_sigma

from astropop.testing import *

def gen_bkg(size, level, rdnoise, rng_seed=123, dtype='f8'):
    """Generate a simple background image."""
    # create a level image
    im = np.ones(size[::-1], dtype)*level

    # reate the gaussian read noise image to sum
    with NumpyRNGContext(rng_seed):
        noise = np.random.normal(loc=0, scale=rdnoise, size=size[::-1])
    im += noise

    # poisonic not needed?
    return im


def gen_position_flux(size, number, low, high, rng_seed=123):
    """Generate x, y, and flux lists for stars."""
    for i in range(number):
        with NumpyRNGContext(rng_seed):
            x = np.random.randint(0, size[0], number)
        with NumpyRNGContext(rng_seed+i):
            y = np.random.randint(0, size[1], number)
            flux = np.random.randint(low, high, number)
    return x, y, flux


def gen_stars_moffat(size, x, y, flux, fwhm):
    """Generate stars image to add to background."""
    beta = 1.5
    alpha = fwhm/np.sqrt(2**(1/beta)-1)
    im = np.zeros(size[::-1])
    grid_y, grid_x = np.indices(size[::-1])
    for xi, yi, fi in zip(x, y, flux):
        imi, gxi, gyi = trim_array(np.zeros_like(im), box_size=5*fwhm,
                                   position=(xi, yi),
                                   indices=(grid_y, grid_x))
        imi += moffat_2d(gxi, gyi, xi, yi, alpha, beta, fi, 0)
        im[gyi.min():gyi.max()+1, gxi.min():gxi.max()+1] += imi

    return im


def gen_stars_gaussian(size, x, y, flux, sigma, theta):
    """Generate stars image to add to background."""
    im = np.zeros(size[::-1])
    grid_y, grid_x = np.indices(size[::-1])

    try:
        sigma_x, sigma_y = sigma
    except (TypeError, ValueError):
        sigma_x = sigma_y = sigma

    bc = np.broadcast(x, y, flux, sigma_x, sigma_y, theta)

    for xi, yi, fi, sxi, syi, ti in zip(*bc.iters):
        imi, gxi, gyi = trim_array(np.zeros_like(im), box_size=10*sxi,
                                   position=(xi, yi),
                                   indices=(grid_y, grid_x))
        imi += gaussian_2d(gxi, gyi, xi, yi, sxi, syi, ti, fi, 0)
        im[gyi.min():gyi.max()+1, gxi.min():gxi.max()+1] += imi

    return im


def gen_image(size, x, y, flux, sky, rdnoise, model='gaussian', **kwargs):
    """Generate a full image of stars with noise."""
    im = gen_bkg(size, sky, rdnoise)

    if model == 'moffat':
        fwhm = kwargs.pop('fwhm')
        im += gen_stars_moffat(size, x, y, flux, fwhm)
    if model == 'gaussian':
        sigma = kwargs.pop('sigma', 2.0)
        theta = kwargs.pop('theta', 0)
        im += gen_stars_gaussian(size, x, y, flux, sigma, theta)

    # can pass the poisson noise
    if not kwargs.get('skip_poisson', False):
        # prevent negative number error
        negatives = np.where(im < 0)
        im = np.random.poisson(np.absolute(im))
        # restore the negatives
        im[negatives] = -im[negatives]
    return im


class Test_Detection_Conformance():
    # Conformance of secondary functions.
    def test_gen_kernel(self):
        assert_equal(gen_filter_kernel(3), [[0, 1, 0],
                                            [1, 4, 1],
                                            [0, 1, 0]])

        assert_equal(gen_filter_kernel(5), [[1, 2, 3, 2, 1],
                                            [2, 4, 6, 4, 2],
                                            [3, 6, 9, 6, 3],
                                            [2, 4, 6, 4, 2],
                                            [1, 2, 3, 2, 1]])

        assert_equal(gen_filter_kernel(7), [[1, 2, 3, 4, 3, 2, 1],
                                            [2, 4, 6, 8, 6, 4, 2],
                                            [3, 6, 9, 12, 9, 6, 3],
                                            [4, 8, 12, 16, 12, 8, 4],
                                            [3, 6, 9, 12, 9, 6, 3],
                                            [2, 4, 6, 8, 6, 4, 2],
                                            [1, 2, 3, 4, 3, 2, 1]])

    @pytest.mark.skip
    def test_sources_mask(self):
        pass


class Test_SepFixByteOrder():
    def test_sep_fix_byte_order_memmaparray(self):
        arr = MemMapArray(np.arange(10).reshape(2, 5))
        arr1 = _sep_fix_byte_order(arr)
        assert_equal(arr1, arr)
        assert_is_instance(arr1, np.ndarray)
        assert_true(arr1.dtype.isnative)
        assert_true(arr1.flags['C_CONTIGUOUS'])
        assert_equal(arr1.dtype, np.dtype('float64'))

    def test_sep_fix_byte_order_f4(self):
        arr = np.arange(10, dtype='f4').reshape(2, 5)
        arr1 = _sep_fix_byte_order(arr)
        assert_equal(arr1, arr)
        assert_is_instance(arr1, np.ndarray)
        assert_true(arr1.dtype.isnative)
        assert_true(arr1.flags['C_CONTIGUOUS'])
        assert_equal(arr1.dtype, np.dtype('float32'))

    @pytest.mark.parametrize('dtype', ['i1', 'i2', 'i4', 'i8',
                                       'u1', 'u2', 'u4', 'u8'])
    def test_sep_fix_byte_order_dtypes(self, dtype):
        arr = np.arange(10, dtype=dtype).reshape(2, 5)
        arr1 = _sep_fix_byte_order(arr)
        assert_equal(arr1, arr)
        assert_is_instance(arr1, np.ndarray)
        assert_true(arr1.dtype.isnative)
        assert_true(arr1.flags['C_CONTIGUOUS'])
        assert_equal(arr1.dtype, np.dtype('float64'))

    @pytest.mark.parametrize('byteorder', ['>', '<'])
    @pytest.mark.parametrize('dtype', ['i1', 'i2', 'i4', 'i8',
                                       'u1', 'u2', 'u4', 'u8',
                                       'f8'])
    def test_sep_fix_byte_order_byteorder(self, byteorder, dtype):
        dtype = f"{byteorder}{dtype}"
        arr = np.arange(10, dtype=dtype).reshape(2, 5)
        arr1 = _sep_fix_byte_order(arr)
        assert_equal(arr1, arr)
        assert_is_instance(arr1, np.ndarray)
        assert_true(arr1.dtype.isnative)
        assert_true(arr1.flags['C_CONTIGUOUS'])
        assert_equal(arr1.dtype, np.dtype('float64'))

    def test_sep_fix_byte_order_contiguous(self):
        arr = np.arange(10, dtype='f4').reshape(2, 5).T
        arr1 = _sep_fix_byte_order(arr)
        assert_equal(arr1, arr)
        assert_is_instance(arr1, np.ndarray)
        assert_true(arr1.dtype.isnative)
        assert_true(arr1.flags['C_CONTIGUOUS'])
        assert_equal(arr1.dtype, np.dtype('float32'))


class Test_Background():
    def test_background_simple_nocosmic(self):
        size = (2048, 2048)
        level = 800
        rdnoise = 20
        image_test = gen_bkg(size, level, rdnoise)

        box_size = 64
        filter_size = 3
        global_bkg, global_rms = background(image_test, box_size, filter_size,
                                            mask=None, global_bkg=True)
        bkg, rms = background(image_test, box_size, filter_size,
                              mask=None, global_bkg=False)

        assert_equal(type(global_bkg), float)
        assert_equal(type(global_rms), float)
        assert_almost_equal(global_bkg, level, decimal=0)
        assert_almost_equal(global_rms, rdnoise, decimal=0)

        assert_equal(bkg.shape, size)
        assert_equal(rms.shape, size)
        assert_almost_equal(bkg, np.ones(size)*level, decimal=0)
        assert_almost_equal(rms, np.ones(size)*rdnoise, decimal=0)

    def test_background_simple_negative_sky(self):
        size = (2048, 2048)
        level = -100
        rdnoise = 20
        image_test = gen_bkg(size, level, rdnoise)

        box_size = 64
        filter_size = 3
        global_bkg, global_rms = background(image_test, box_size, filter_size,
                                            mask=None, global_bkg=True)
        bkg, rms = background(image_test, box_size, filter_size,
                              mask=None, global_bkg=False)

        assert_equal(type(global_bkg), float)
        assert_equal(type(global_rms), float)
        assert_almost_equal(global_bkg, level, decimal=0)
        assert_almost_equal(global_rms, rdnoise, decimal=0)

        assert_equal(bkg.shape, size)
        assert_equal(rms.shape, size)
        assert_almost_equal(bkg, np.ones(size)*level, decimal=0)
        assert_almost_equal(rms, np.ones(size)*rdnoise, decimal=0)

    def test_background_simple_cosmic(self):
        size = (2048, 2048)
        level = 800
        rdnoise = 20
        image_test = gen_bkg(size, level, rdnoise)

        # add some cosmics
        for i in range(100):  # 100 single pixel cosmics
            x = np.random.randint(0, size[0]-1)
            y = np.random.randint(0, size[1]-1)
            image_test[x, y] = np.random.randint(16000, 64000)

        for i in range(4):  # 4 square block cosmics
            x = np.random.randint(0, size[0]-3)
            y = np.random.randint(0, size[1]-3)
            image_test[x:x+2, y:y+2] = np.random.randint(16000, 64000)

        box_size = 64
        filter_size = 3
        global_bkg, global_rms = background(image_test, box_size, filter_size,
                                            mask=None, global_bkg=True)
        bkg, rms = background(image_test, box_size, filter_size,
                              mask=None, global_bkg=False)

        assert_equal(type(global_bkg), float)
        assert_equal(type(global_rms), float)
        assert_almost_equal(global_bkg, level, decimal=0)
        assert_almost_equal(global_rms, rdnoise, decimal=0)

        assert_equal(bkg.shape, size)
        assert_equal(rms.shape, size)
        assert_almost_equal(bkg, np.ones(size)*level, decimal=0)
        assert_almost_equal(rms, np.ones(size)*rdnoise, decimal=0)

    def test_background_variablelevel_cosmics(self):
        size = (1024, 1024)
        y_i, x_i = np.indices(size)
        level = x_i*y_i/500  # level from 0 to 2000
        rdnoise = 20
        image_test = gen_bkg(size, level, rdnoise)

        # add some cosmics
        for i in range(100):  # 100 single pixel cosmics
            x = np.random.randint(0, size[0]-1)
            y = np.random.randint(0, size[1]-1)
            image_test[x, y] = np.random.randint(16000, 64000)

        for i in range(4):  # 4 square block cosmics
            x = np.random.randint(0, size[0]-3)
            y = np.random.randint(0, size[1]-3)
            image_test[x:x+2, y:y+2] = np.random.randint(16000, 64000)

        box_size = 64
        filter_size = 3
        global_bkg, global_rms = background(image_test, box_size, filter_size,
                                            mask=None, global_bkg=True)
        bkg, rms = background(image_test, box_size, filter_size,
                              mask=None, global_bkg=False)

        assert_equal(type(global_bkg), float)
        assert_equal(type(global_rms), float)
        assert_almost_equal(global_bkg, 389, decimal=0)
        assert_almost_equal(global_rms, 36, decimal=0)

        assert_equal(bkg.shape, size)
        assert_equal(rms.shape, size)
        # with stars, the dispersion increases
        assert_almost_equal(bkg, np.ones(size)*level, decimal=-1)
        # assert_almost_equal(rms, np.ones(size)*rdnoise, decimal=-1)

    def test_background_stars(self):
        size = (1024, 1024)
        stars_n = 50
        flux_low = 1500
        flux_high = 25000
        fwhm = 5
        level = 800
        rdnoise = 20
        x, y, f = gen_position_flux(size, stars_n, flux_low, flux_high)
        image_test = gen_bkg(size, level, rdnoise)
        image_test += gen_stars_moffat(size, x, y, f, fwhm)

        box_size = 64
        filter_size = 3
        global_bkg, global_rms = background(image_test, box_size, filter_size,
                                            mask=None, global_bkg=True)
        bkg, rms = background(image_test, box_size, filter_size,
                              mask=None, global_bkg=False)

        assert_equal(type(global_bkg), float)
        assert_equal(type(global_rms), float)
        assert_almost_equal(global_bkg, level, decimal=0)
        assert_almost_equal(global_rms, rdnoise, decimal=0)

        assert_equal(bkg.shape, size)
        assert_equal(rms.shape, size)
        # with stars, the dispersion increases
        assert_almost_equal(bkg, np.ones(size)*level, decimal=-1)
        assert_almost_equal(rms, np.ones(size)*rdnoise, decimal=-1)

    def test_background_stars_variablelevel(self):
        size = (1024, 1024)
        stars_n = 50
        flux_low = 1500
        flux_high = 25000
        fwhm = 5
        y_i, x_i = np.indices(size)
        level = x_i*y_i/500  # level from 0 to 2000
        rdnoise = 20
        x, y, f = gen_position_flux(size, stars_n, flux_low, flux_high)
        image_test = gen_bkg(size, level, rdnoise)
        image_test += gen_stars_moffat(size, x, y, f, fwhm)

        box_size = 64
        filter_size = 3
        global_bkg, global_rms = background(image_test, box_size, filter_size,
                                            mask=None, global_bkg=True)
        bkg, rms = background(image_test, box_size, filter_size,
                              mask=None, global_bkg=False)

        assert_equal(type(global_bkg), float)
        assert_equal(type(global_rms), float)
        assert_almost_equal(global_bkg, 389, decimal=0)
        assert_almost_equal(global_rms, 36, decimal=0)

        assert_equal(bkg.shape, size)
        assert_equal(rms.shape, size)
        # with stars, the dispersion increases
        assert_almost_equal(bkg, np.ones(size)*level, decimal=-1)
        # here the rms goes bigger with the level
        # assert_almost_equal(rms, np.ones(size)*rdnoise, decimal=-1)


@pytest.mark.flaky(max_runs=10, min_passes=1)
class Test_SEP_Detection():
    # segmentation detection. Must detect all shapes of sources

    def resort_sources(self, x, y, f):
        """Random sources are random. We must resort to compare."""
        # For SEP, resort using x order
        order = np.argsort(y)
        return x[order], y[order], f[order]

    def test_sepfind_one_star(self):
        size = (128, 128)
        pos = (64, 64)
        sky = 20
        sky = 800
        rdnoise = 20
        flux = 32000
        sigma = 3
        theta = 0
        threshold = 10
        im = gen_image(size, [pos[0]], [pos[1]], [flux], sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = sepfind(im, threshold, sky, rdnoise)

        assert_equal(len(sources), 1)
        assert_almost_equal(sources['x'][0], 64, decimal=0)
        assert_almost_equal(sources['y'][0], 64, decimal=0)

    def test_sepfind_negative_sky(self):
        size = (128, 128)
        pos = (64, 64)
        sky = 0
        rdnoise = 20
        flux = 32000
        sigma = 3
        theta = 0
        threshold = 10

        im = gen_image(size, [pos[0]], [pos[1]], [flux], sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = sepfind(im, threshold, sky, rdnoise)

        assert_equal(len(sources), 1)
        assert_almost_equal(sources['x'][0], 64, decimal=0)
        assert_almost_equal(sources['y'][0], 64, decimal=0)

    def test_sepfind_strong_and_weak(self):
        size = (128, 128)
        posx = (60, 90)
        posy = (20, 90)
        sky = 800
        rdnoise = 20
        flux = (32000, 600)
        sigma = 1.5
        theta = 0
        im = gen_image(size, posx, posy, flux, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = sepfind(im, 3, sky, rdnoise)

        assert_equal(len(sources), 2)
        assert_almost_equal(sources['x'], posx, decimal=0)
        assert_almost_equal(sources['y'], posy, decimal=0)

    def test_sepfind_four_stars_fixed_position(self):
        size = (1024, 1024)
        posx = (10, 120, 500, 1000)
        posy = (20, 200, 600, 800)
        sky = 800
        rdnoise = 20
        flux = (15000, 1500, 1500, 35000)
        sigma = 3
        theta = 0
        im = gen_image(size, posx, posy, flux, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = sepfind(im, 3, sky, rdnoise)

        assert_equal(len(sources), 4)
        assert_almost_equal(sources['x'], posx, decimal=0)
        assert_almost_equal(sources['y'], posy, decimal=0)

    def test_sepfind_multiple_stars(self):
        size = (1024, 1024)
        number = 15
        low = 2000
        high = 30000
        sky = 800
        rdnoise = 20
        sigma = 3
        theta = 0

        x, y, f = gen_position_flux(size, number, low, high, rng_seed=456)
        im = gen_image(size, x, y, f, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = sepfind(im, 5, sky, rdnoise)

        x, y, f = self.resort_sources(x, y, f)

        assert_equal(len(sources), number)
        assert_almost_equal(sources['x'], x, decimal=0)
        assert_almost_equal(sources['y'], y, decimal=0)

    def test_sepfind_one_star_subpixel(self):
        size = (128, 128)
        pos = (54.32, 47.86)

        im = gen_image(size, [pos[0]], [pos[1]], [45000], 800, 0,
                       sigma=[3], skip_poisson=True)
        sources = sepfind(im, 5, 800, 10)
        assert_equal(len(sources), 1)
        # no error, 2 decimals ok!
        assert_almost_equal(sources[0]['x'], pos[0], decimal=2)
        assert_almost_equal(sources[0]['y'], pos[1], decimal=2)


@pytest.mark.flaky(max_runs=10, min_passes=1)
class Test_DAOFind_Detection():
    # DAOFind detection. Only round unsturaded stars

    def resort_sources(self, x, y, f):
        """Random sources are random. We must resort to compare."""
        # For SEP, resort using x order
        order = np.argsort(y)
        return x[order], y[order], f[order]

    @pytest.mark.parametrize('fwhm', np.arange(1, 9, 1))
    def test_daofind_constants_calc(self, fwhm):
        # Compare our constants with D.Jones daofind
        dao = DAOFind(fwhm)

        maxbox = 13  # Maximum size of convolution box in pixels
        assert_equal(dao._maxbox, maxbox)
        radius = np.max([0.637*fwhm, 2.001])
        assert_equal(dao._radius, radius)
        radsq = radius**2
        nhalf = np.min([int(radius), int((maxbox-1)/2.)])
        assert_equal(dao._nhalf, nhalf)
        nbox = 2*nhalf + 1  # number of pixels in side of convolution box
        assert_equal(dao._nbox, nbox)

        sigsq = (fwhm*gaussian_fwhm_to_sigma)**2
        assert_equal(dao._sigma2, sigsq)

        mask = np.zeros([nbox, nbox], dtype='int8')
        g = np.zeros([nbox, nbox])  # Gaussian convolution kernel
        row2 = (np.arange(nbox)-nhalf)**2
        for i in range(nhalf+1):
            temp = row2 + i**2
            g[nhalf-i, :] = temp
            g[nhalf+i, :] = temp
        g_row = np.where(g <= radsq)
        # MASK is complementary to SKIP in Stetson's Fortran
        mask[g_row[0], g_row[1]] = 1
        assert_equal(dao._conv_mask, mask)
        good = np.where(mask)  # Value of c are now equal to distance to center
        pixels = len(good[0])

        # Compute quantities for centroid computations that can be used for all
        # stars
        g = np.exp(-0.5*g/sigsq)
        assert_equal(dao._g, g)

        xwt = np.zeros([nbox, nbox])
        wt = nhalf - np.abs(np.arange(nbox)-nhalf) + 1
        assert_equal(dao._wt, wt)
        for i in range(nbox):
            xwt[i, :] = wt
        assert_equal(dao._xwt, xwt)
        ywt = np.transpose(xwt)
        assert_equal(dao._ywt, ywt)
        sgx = np.sum(g*xwt, 1)
        assert_equal(dao._sgx, sgx)
        p = np.sum(wt)
        assert_equal(dao._p, p)
        sgy = np.sum(g*ywt, 0)
        assert_equal(dao._sgy, sgy)
        sumgx = np.sum(wt*sgy)
        assert_equal(dao._sumgx, sumgx)
        sumgy = np.sum(wt*sgx)
        assert_equal(dao._sumgy, sumgy)
        sumgsqy = np.sum(wt*sgy*sgy)
        assert_equal(dao._sumgsqy, sumgsqy)
        sumgsqx = np.sum(wt*sgx*sgx)
        assert_equal(dao._sumgsqx, sumgsqx)
        vec = nhalf - np.arange(nbox)
        assert_equal(dao._vec, vec)
        dgdx = sgy*vec
        assert_equal(dao._dgdx, dgdx)
        dgdy = sgx*vec
        assert_equal(dao._dgdy, dgdy)
        sdgdxs = np.sum(wt*dgdx**2)
        assert_equal(dao._sdgdxs, sdgdxs)
        sdgdx = np.sum(wt*dgdx)
        assert_equal(dao._sdgdx, sdgdx)
        sdgdys = np.sum(wt*dgdy**2)
        assert_equal(dao._sdgdys, sdgdys)
        sdgdy = np.sum(wt*dgdy)
        assert_equal(dao._sdgdy, sdgdy)
        sgdgdx = np.sum(wt*sgy*dgdx)
        assert_equal(dao._sgdgdx, sgdgdx)
        sgdgdy = np.sum(wt*sgx*dgdy)
        assert_equal(dao._sgdgdy, sgdgdy)

        c = g*mask  # Convolution kernel now in c
        sumc = np.sum(c)
        sumcsq = np.sum(c**2) - sumc**2/pixels
        sumc = sumc/pixels
        c[good[0], good[1]] = (c[good[0], good[1]] - sumc)/sumcsq
        assert_equal(dao._kernel, c)
        c1 = np.exp(-.5*row2/sigsq)
        sumc1 = np.sum(c1)/nbox
        sumc1sq = np.sum(c1**2) - sumc1
        c1 = (c1-sumc1)/sumc1sq
        assert_equal(dao._c1, c1)

    def test_daofind_sharpness(self):
        # compare my implementation with D.Jones PythonPhot
        # without filtering, both have to output the same round/sharp and
        # the same coordinates for all stars, because use the same algorithm

        image_size = (525, 200)
        xpos = np.arange(10)*50 + 25
        ypos = np.ones_like(xpos)*30 + np.arange(len(xpos))*10
        sky = 800
        rdnoise = 50
        threshold = 50

        fwhm = 5
        sigma = np.array([0.1, 0.5, 0.8, 1.0, 1.2, 1.5, 2.0, 3.5, 3.0, 3.5])
        sigma *= fwhm*gaussian_fwhm_to_sigma
        flux = (np.ones_like(xpos)*sigma)*80000
        expect_sharp = [3.55, 0.76, 0.50, 0.45, 0.43,
                        0.41, 0.39, 0.39, 0.39, 0.39]

        im = gen_image(image_size, xpos, ypos, flux, sky, rdnoise,
                       model='gaussian', sigma=sigma)

        sources = daofind(im, threshold, sky, rdnoise, fwhm,
                          sharp_limit=None, round_limit=None)

        assert_equal(len(sources), len(xpos))
        assert_almost_equal(sources['x'], xpos, decimal=1)
        assert_almost_equal(sources['y'], ypos, decimal=1)
        assert_almost_equal(sources['sharp'], expect_sharp, decimal=1)

    def test_daofind_roundness(self):
        # compare my implementation with D.Jones PythonPhot
        # without filtering, both have to output the same round/sharp and
        # the same coordinates for all stars, because use the same algorithm

        image_size = (525, 200)
        xpos = np.arange(10)*50 + 25
        ypos = np.ones_like(xpos)*30 + np.arange(len(xpos))*10
        sky = 800
        rdnoise = 50
        threshold = 50

        e = (np.arange(len(xpos)))*0.1
        flux = np.sqrt(np.arange(len(xpos))+1)*80000
        sigma_x = np.sqrt(1/(1-e**2))*2
        sigma_y = np.ones_like(xpos)*2
        fwhm = 5
        theta = np.zeros_like(xpos)
        sky = 800
        rdnoise = 20  # very low noise
        expect_round = np.array([0.0, 0.02, 0.03, 0.08, 0.15, 0.25,
                                 0.39, 0.59, 0.88, 1.39])

        im = gen_image(image_size, xpos, ypos, flux, sky, rdnoise,
                       model='gaussian', sigma=(sigma_x, sigma_y),
                       theta=theta)

        sources = daofind(im, threshold, sky, rdnoise, fwhm,
                          sharp_limit=None, round_limit=None)

        assert_equal(len(sources), len(xpos))
        assert_almost_equal(sources['x'], xpos, decimal=0)
        assert_almost_equal(sources['y'], ypos, decimal=0)
        assert_almost_equal(sources['round'], -expect_round, decimal=1)

    def test_daofind_one_star(self):
        size = (128, 128)
        pos = (64, 64)
        sky = 70
        rdnoise = 20
        flux = 32000
        theta = 0
        fwhm = 3
        sigma = fwhm*gaussian_fwhm_to_sigma
        threshold = 10

        im = gen_image(size, [pos[0]], [pos[1]], [flux], sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = daofind(im, threshold, sky, rdnoise, fwhm)

        assert_equal(len(sources), 1)
        assert_almost_equal(sources['x'][0], pos[0], decimal=1)
        assert_almost_equal(sources['y'][0], pos[1], decimal=1)

    def test_daofind_strong_and_weak(self):
        size = (128, 128)
        posx = (45, 90)
        posy = (45, 90)
        sky = 800
        rdnoise = 20
        flux = (32000, 16000)
        fwhm = 3
        sigma = fwhm*gaussian_fwhm_to_sigma
        theta = 0
        im = gen_image(size, posx, posy, flux, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)
        threshold = 10

        sources = daofind(im, threshold, sky, rdnoise, fwhm)

        assert_equal(len(sources), 2)
        assert_almost_equal(sources['x'], posx, decimal=0)
        assert_almost_equal(sources['y'], posy, decimal=0)

    def test_daofind_four_stars_fixed_position(self):
        size = (128, 128)
        posx = (45, 90, 45, 90)
        posy = (45, 50, 90, 100)
        sky = 800
        rdnoise = 20
        flux = (15000, 3000, 5000, 35000)
        fwhm = 3
        sigma = fwhm*gaussian_fwhm_to_sigma
        theta = 0
        threshold = 10

        im = gen_image(size, posx, posy, flux, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = daofind(im, threshold, sky, rdnoise, fwhm)

        assert_equal(len(sources), 4)
        assert_almost_equal(sources['x'], posx, decimal=0)
        assert_almost_equal(sources['y'], posy, decimal=0)

    def test_daofind_multiple_stars(self):
        size = (512, 512)
        number = 15
        low = 2000
        high = 30000
        sky = 800
        rdnoise = 20
        fwhm = 5
        sigma = fwhm*gaussian_fwhm_to_sigma
        theta = 0
        threshold = 10
        x, y, f = gen_position_flux(size, number, low, high, rng_seed=456)

        im = gen_image(size, x, y, f, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = daofind(im, threshold, sky, rdnoise, fwhm)

        x, y, f = self.resort_sources(x, y, f)

        assert_equal(len(sources), number)
        assert_almost_equal(sources['x'], x, decimal=0)
        assert_almost_equal(sources['y'], y, decimal=0)

    def test_daofind_reject_roundness(self):
        size = (128, 128)
        pos_x = [20, 60, 100, 40, 80]
        pos_y = [20, 30, 40, 50, 60]
        sky = 70
        rdnoise = 20
        flux = [30000]*5
        theta = 0
        fwhm = 3
        sigma_x = np.array([1, 0.5, 1, 2.0, 0.1])*gaussian_fwhm_to_sigma*fwhm
        sigma_y = np.array([1, 1.0, 1, 0.5, 0.1])*gaussian_fwhm_to_sigma*fwhm
        threshold = 10
        # stars 0, 2 -> passed
        # star 4 -> rejected by sharpness
        # stars 1, 3 -> rejected by roundness

        im = gen_image(size, pos_x, pos_y, flux, sky, rdnoise,
                       model='gaussian', sigma=(sigma_x, sigma_y),
                       theta=theta)

        sources = daofind(im, threshold, sky, rdnoise, fwhm,
                          sharp_limit=(0.3, 0.6), round_limit=(-0.5, 0.5))

        assert_equal(len(sources), 2)
        assert_almost_equal(sources['x'], [20, 100], decimal=0)
        assert_almost_equal(sources['y'], [20, 40], decimal=0)

    def test_daofind_one_star_subpixel(self):
        size = (128, 128)
        pos = (54.32, 47.86)

        im = gen_image(size, [pos[0]], [pos[1]], [45000], 800, 0,
                       sigma=[5*gaussian_fwhm_to_sigma],
                       skip_poisson=True)
        sources = daofind(im, 5, 800, 10, 5)
        assert_equal(len(sources), 1)
        # no error, 2 decimals ok!
        assert_almost_equal(sources[0]['x'], pos[0], decimal=2)
        assert_almost_equal(sources[0]['y'], pos[1], decimal=2)


@pytest.mark.flaky(max_runs=10, min_passes=1)
class Test_StarFind():
    # Our combined iterative method

    def resort_sources(self, x, y, f):
        """Random sources are random. We must resort to compare."""
        # For SEP, resort using x order
        order = np.argsort(y)
        return x[order], y[order], f[order]

    def test_starfind_calc_fwhm(self):
        size = (512, 512)
        number = 15
        sky = 70
        rdnoise = 20
        low = 120000
        high = 320000
        sigma = 5
        theta = 0

        x, y, f = gen_position_flux(size, number, low, high, rng_seed=456)

        im = gen_image(size, x, y, f, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        fwhm = calc_fwhm(im, x, y, box_size=25, model='gaussian', min_fwhm=3.0)
        assert_almost_equal(fwhm, 2.35*sigma, decimal=0)

    pytest.mark.skip('Wrong new centers?')
    def test_starfind_recenter_sources(self):
        size = (256, 256)
        number = 10
        sky = 70
        rdnoise = 20
        low = 120000
        high = 320000
        sigma = 5
        theta = 0

        x, y, f = gen_position_flux(size, number, low, high, rng_seed=456)

        im = gen_image(size, x, y, f, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        nx, ny = recenter_sources(im,
                                  x+np.random.normal(loc=0, scale=1,
                                                     size=len(x)),
                                  y+np.random.normal(loc=0, scale=1,
                                                     size=len(x)),
                                  box_size=15, model='gaussian')

        assert_equal(len(nx), number)
        assert_equal(len(ny), number)
        # TODO: this seems unprecise. Investigate it.
        assert_almost_equal(nx, x, decimal=-1)
        assert_almost_equal(ny, y, decimal=-1)

    def test_starfind_one_star(self):
        size = (128, 128)
        x, y = (64, 64)
        f = 80000
        sky = 70
        rdnoise = 20
        sigma = 5
        theta = 0
        fwhm = 5  # dummy low value
        threshold = 10

        im = gen_image(size, [x], [y], [f], sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = starfind(im, threshold, sky, rdnoise, fwhm)

        assert_equal(len(sources), 1)
        assert_almost_equal(sources['x'], x, decimal=0)
        assert_almost_equal(sources['y'], y, decimal=0)
        assert_almost_equal(sources.meta['astropop fwhm'], sigma*2.355,
                            decimal=1)

    def test_starfind_strong_weak(self):
        size = (200, 100)
        posx = (50, 150)
        posy = (40, 60)
        sky = 800
        rdnoise = 20
        flux = (64000, 6000)
        theta = 0
        fwhm = 3
        sigma = fwhm*gaussian_fwhm_to_sigma
        threshold = 10
        im = gen_image(size, posx, posy, flux, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = starfind(im, threshold, sky, rdnoise, fwhm)

        assert_equal(len(sources), 2)
        assert_almost_equal(sources['x'], posx, decimal=1)
        assert_almost_equal(sources['y'], posy, decimal=1)
        assert_almost_equal(sources.meta['astropop fwhm'], sigma*2.355,
                            decimal=1)

    def test_starfind_blind_fwhm(self):
        size = (512, 512)
        number = 12
        low, high = (15000, 60000)
        sky = 800
        rdnoise = 20
        sigma = 2
        theta = 0
        threshold = 8

        x, y, f = gen_position_flux(size, number, low, high, rng_seed=456)
        x, y, f = self.resort_sources(x, y, f)

        im = gen_image(size, x, y, f, sky, rdnoise,
                       model='gaussian', sigma=sigma, theta=theta)

        sources = starfind(im, threshold, sky, rdnoise)

        assert_equal(len(sources), number)
        assert_almost_equal(sources['x'], x, decimal=1)
        assert_almost_equal(sources['y'], y, decimal=1)
        assert_almost_equal(sources.meta['astropop fwhm'], sigma*2.355,
                            decimal=2)

    def test_starfind_rejection(self):
        size = (128, 128)
        pos_x = [20, 60, 100, 40, 80, 50]
        pos_y = [20, 30, 40, 50, 60, 85]
        sky = 70
        rdnoise = 20
        flux = [30000]*6
        theta = 0
        fwhm = 3
        sig_base = gaussian_fwhm_to_sigma*fwhm
        sigma_x = np.array([1, 0.5, 1, 2.0, 0.1, 1])*sig_base
        sigma_y = np.array([1, 1.0, 1, 0.5, 0.1, 1])*sig_base
        threshold = 10
        # stars 0, 2 -> passed
        # star 4 -> rejected by sharpness
        # stars 1, 3 -> rejected by roundness

        im = gen_image(size, pos_x, pos_y, flux, sky, rdnoise,
                       model='gaussian', sigma=(sigma_x, sigma_y),
                       theta=theta)

        sources = starfind(im, threshold, sky, rdnoise,
                           sharp_limit=(0.3, 0.6), round_limit=(-0.5, 0.5))

        assert_equal(len(sources), 3)
        assert_almost_equal(sources['x'], [20, 100, 50], decimal=0)
        assert_almost_equal(sources['y'], [20, 40, 85], decimal=0)
        assert_almost_equal(sources.meta['astropop fwhm'], fwhm,
                            decimal=1)

    def test_starfind_one_star_subpixel(self):
        size = (128, 128)
        pos = (54.32, 47.86)

        im = gen_image(size, [pos[0]], [pos[1]], [45000], 800, 0,
                       sigma=[5*gaussian_fwhm_to_sigma],
                       skip_poisson=True)
        sources = starfind(im, 5, 800, 10)
        assert_equal(len(sources), 1)
        # no error, 2 decimals ok!
        assert_almost_equal(sources[0]['x'], pos[0], decimal=2)
        assert_almost_equal(sources[0]['y'], pos[1], decimal=2)
